"""
This module contains the core functionality of larch.reactive.
"""
import sys
import warnings
import logging
import operator
from bisect import insort_left
from collections import deque
from functools import reduce, update_wrapper
from weakref import ref

pointer_itemgetter = operator.itemgetter
pointer_attrgetter = operator.attrgetter


# __pragma__ ('ecom')
"""?
class LocalContext:
    pass


class GreenletExit(BaseException):
    pass
?"""

# __pragma__ ('skip')
pointer_attrgetter = type(pointer_attrgetter(""))  # for pypy
try:
    from gevent import GreenletExit
    from gevent.local import local as LocalContext

except ImportError:  # pragma: no cover
    class GreenletExit(BaseException):
        pass

    from threading import local as LocalContext

__all__ = ("rcontext", "CellBase", "ReactiveState", "Subject", "Container",
           "ResetContainer", "Rule", "IterRule", "pointer_attrgetter",
           "pointer_itemgetter", "pointer_resolve", "ReactiveWarning")
# __pragma__ ('noskip')

logger = logging.getLogger("larch.reactive")


def pointer_resolve(root, accessors):
    return reduce(lambda x, y: y(x), accessors, root)


class DecoratorContext:
    """
    A base class for objects that can be used
    as decorators and as contexts.
    """
    __slots__ = ()

    # __pragma__("kwargs")
    def call(self, func, *args, **kwargs):
        with self:
            result = func(*args, **kwargs)
        return result  # transcript bug
    # __pragma__("nokwargs")

    def __call__(self, func):
        # __pragma__("kwargs")
        def handler(*args, **kwargs):
            return self.call(func, *args, **kwargs)
        # __pragma__("nokwargs")

        result = update_wrapper(handler, func)
        result.decorated = func
        return result


class AtomicDecoratorContext(DecoratorContext):
    """
    The atomic decorator context. Rules will be executed at the end of
    the context.
    """

    def __enter__(self):
        rcontext._atomic_start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        rcontext._atomic_end()
        return False


class SilentDecoratorContext(DecoratorContext):
    """
    The silent decorator context. Within this context a change of a cell
    will not trigger any rules.
    """
    __slots__ = ("old_notify", )

    def __enter__(self):
        # the atomic context ensures that silent will not influence
        # other greenlets
        rcontext._atomic_start()
        self.old_notify = rcontext.notify
        rcontext.notify = rcontext._dumy_notify

    def __exit__(self, exc_type, exc_val, exc_tb):
        rcontext.notify = self.old_notify
        rcontext._atomic_end()
        return False


class UntouchedDecoratorContext(DecoratorContext):
    """
    The untouched decorator context. This context makes only sense within
    rules. Within this context no cell will be touched.
    """
    __slots__ = ("old_touch", )

    def __enter__(self):
        self.old_touch = rcontext.touch
        rcontext.touch = rcontext._dumy_touch

    def __exit__(self, exc_type, exc_val, exc_tb):
        rcontext.touch = self.old_touch
        return False


class TouchedDecoratorContext(DecoratorContext):
    """
    The touch decorator context. The context makes only sense within
    a untouched context. Within this context cells will be touched again.
    """
    __slots__ = ("old_touch", )

    def __enter__(self):
        self.old_touch = rcontext.touch
        rcontext.touch = rcontext._touch

    def __exit__(self, exc_type, exc_val, exc_tb):
        rcontext.touch = self.old_touch
        return False


class ReactiveWarning(UserWarning):
    pass


class ReactiveContext(LocalContext):
    """
    A ReactiveContext is the core driver for reactive behaviour.
    It manages the subject - observer relation between cells and rules,
    and cares for the execution order of rules.

    Only one ReativeContext exists in a program that can be accessed by:

    >>> larch.reactive.rcontext
    """

    def __init__(self):
        self.atomic_count = 0
        self.atomic_start_round = 0
        self._call_counter = 0
        self.current_observer = None
        self.touch = self._dumy_touch
        self.notify = self._notify

        self.old_values = {}
        # the subject values before an atomic operation.

        self.observers = []
        # all observers in the active atomic operation in their call order

        self.observer_count = {}
        # how often is one observer in the observers list

        self.callbacks = deque()
        # sequence of all callbacks to notify the end of notification process,
        # transformed from set to list to keep the callbacks order.

        self.atomic = AtomicDecoratorContext()

    @property
    def untouched(self):
        """UntouchedDectoratorContext: a new context."""
        return UntouchedDecoratorContext()

    @property
    def touched(self):
        """TouchedDecoratorContext: a new context."""
        return TouchedDecoratorContext()

    @property
    def silent(self):
        """SilentDecoratorContext: a new context."""
        return SilentDecoratorContext()

    @property
    def rounds(self):
        """int: the count of atomic operations, since program start."""
        return self.atomic_start_round

    @property
    def transaction_level(self):
        """int: count of recursive atomic operations."""
        return self.atomic_count

    @property
    def inside_rule(self):
        """bool: `True` if the current frame is inside a rule."""
        return self.current_observer

    @property
    def inside_touch(self):
        """bool: `True` if the current frame is inside a touch context."""
        return self.touch == self._touch

    def _atomic_start(self):
        self.atomic_start_round = self._call_counter
        self.atomic_count += 1

    def _atomic_end(self):
        if self.atomic_count != 1:
            self.atomic_count -= 1
            return

        # self.atomic_count == 1
        observers = self.observers
        observer_count = self.observer_count
        greenlet_exit = False

        # execute rules
        try:
            # a "for o in observers" loop is not possible
            # observer.notify() can change new cells and modify
            # the observers list
            # the following lines correctly run the testcases
            # test_renew_cell and test_one_run in test_reactive.py

            for i in range(10000):  # avoid an infinite loop
                # __pragma__ ('tconv')
                if not observers:
                    observer_count.clear()
                    break
                # __pragma__ ('notconv')

                # observers are sorted in reverse order
                observer = observers.pop()
                key_observer = id(observer)
                if observer_count[key_observer] <= 1:
                    del observer_count[key_observer]
                    try:
                        observer.notify()
                    except GreenletExit:
                        greenlet_exit = True
                        logger.exception(
                            "GreenletExit during observer notification", stack_info=True)
                else:
                    # this observer occurs multiple times. Only call the
                    # latest occurence
                    observer_count[key_observer] -= 1
            else:
                self.callbacks.clear()
                exp = RuntimeError("possible endless recursion {0}".format(observers))
                del observers[:]
                observer_count.clear()
                raise exp
        finally:
            self.atomic_count = 0
            self.active_greenlet = None
            self.old_values.clear()

        greenlet_exit = self._execute_callbacks(greenlet_exit)
        if greenlet_exit:
            raise GreenletExit()

    def _execute_callbacks(self, greenlet_exit):
        callbacks = self.callbacks
        self.callbacks = deque()
        for c in callbacks:
            try:
                c()
            except Exception as e:
                # __pragma__ ('skip')
                exc_type, exc_value, tb = sys.exc_info()
                from .exception import format_full_exception_tb
                tb = format_full_exception_tb()
                msg = 'Exception while executing callback: "{0}"\n{1}'
                warnings.warn(msg.format(exc_value, tb), ReactiveWarning)
                logger.exception("Exception while executing callback %r: %r", c, e, stack_info=True)
                # __pragma__ ('noskip')
                """?
                console.trace("Exception while executing callback", repr(e), repr(c), e.stack)
                ?"""
            except GreenletExit:
                greenlet_exit = True
                logger.exception("GreenletExit while executing callback %r", c, stack_info=True)

        return greenlet_exit

    def rule_call(self, rule, args, kwargs):
        old_touch = self.touch
        old_observer = self.current_observer
        self.current_observer = rule
        self.touch = self._touch
        try:
            rule.__call__(args, kwargs)
        except GreenletExit:
            raise
        except BaseException as e:
            # __pragma__ ('skip')
            exc_type, exc_value, tb = sys.exc_info()
            from .exception import format_full_exception_tb
            tb = format_full_exception_tb()
            msg = 'Exception while calling rule "{0}({1!r}, {2!r})": "{3!r}"\n{4}'
            warnings.warn(msg.format(rule, args, kwargs, exc_value, tb), ReactiveWarning)
            logger.exception("Exception while calling rule %r", e, stack_info=True)
            # __pragma__ ('noskip')
            """?
            console.trace("Exception while calling rule", repr(e), repr(rule), e.stack)
            ?"""

        finally:
            self.touch = old_touch
            self.current_observer = old_observer

    def _notify(self, subject, old_value):
        id_subject = id(subject)
        self.old_values.setdefault(id_subject, old_value)
        self.emit(subject.get_observers())
        subject.clear_observers()

    def _dumy_notify(self, subject, old_value):
        pass

    def emit(self, observers):
        my_observers = self.observers
        my_observer_count = self.observer_count
        for o in observers:
            insort_left(my_observers, o)
            ko = id(o)
            my_observer_count[ko] = my_observer_count.get(ko, 0) + 1

    def push_callback(self, callback):
        self.callbacks.append(callback) if self.transaction_level else callback()

    def shift_callback(self, callback):
        self.callbacks.appendleft(callback) if self.transaction_level else callback()

    def call_counter(self):
        self._call_counter += 1
        return self._call_counter

    def _touch(self, subject):
        subject.add_observer(self.current_observer)
        return True

    def _dumy_touch(self, subject):
        return False


rcontext = ReactiveContext()


class Observer:
    level = 0
    __slots__ = ("last_call", )

    def notify(self, initial=False):
        pass  # pragma: no cover

    def __lt__(self, other):
        # observers are sorted in reverse order!
        return (self.level > other.level
                or self.level == other.level and self.last_call > other.last_call)

    def __eq__(self, other):
        return self.level == other.level and self.last_call == other.last_call


class IterRule(Observer):
    """A Rule represents a `rule` instance method of a reactive class"""
    __slots__ = ("level", "method", "holder", "__weakref__")

    def __init__(self, holder, method, level, last_call):
        self.holder = holder
        self.method = method
        self.level = level
        self.last_call = last_call

    def notify(self, args=(), kwargs={}):
        self.last_call = rcontext.call_counter()
        rcontext.rule_call(self, args, kwargs)

    def __repr__(self):
        if self.holder() is None:
            return (f"<Rule obsolete {self.method.__name__} "
                    f"({self.level}, {self.last_call})>")

        return (f"<Rule {self.method.__name__}({self.level}, "
                f"{self.last_call}) of {self.holder()!r}>")

    def __call__(self, args, kwargs):
        holder = self.holder()
        if holder is not None:
            iterator = self.method(holder, *args, **kwargs)

            def call_iterator():
                try:
                    next(iterator)
                    rcontext.push_callback(call_iterator)
                except StopIteration:
                    pass
                except RuntimeError as e:
                    if e.args != ('generator raised StopIteration',):
                        raise

            call_iterator()  # __: skip
            """?
            if iterator:
                call_iterator()
            ?"""


# __pragma__ ('skip')
class Rule(IterRule):
    def __call__(self, args, kwargs):
        holder = self.holder()
        if holder is not None:
            try:
                self.method(holder, *args, **kwargs)
            except StopIteration:
                pass
# __pragma__ ('noskip')


class Subject:
    """
    An observer's subject
    """

    __slots__ = ("_observers", "_last_counter")

    def __init__(self):
        self._observers = {}
        """
        A set of observers that will be called when the subjects value changes.
        """

        self._last_counter = 0
        """
        The variable is used to check if a observer still depends on this
        subject if observer.last_call > self._last_counter ==> the observer
        does not depend anymore.
        (see TestReactive.test_decouple of test_reactive.py)
        """

    def touch(self):
        rcontext.touch(self)

    def changed(self, old_value=None):
        # __pragma__ ('tconv')
        if self._observers:
            rcontext.atomic.call(rcontext.notify, self, old_value)
        # __pragma__ ('notconv')

    def take(self, other):
        """takes over the attributes of another subject"""
        self._observers.update(other._observers)
        self._last_counter = max(self._last_counter, other._last_counter)

    def get_observers(self):
        """Returns the subject's observers"""
        return self.filter_observers(self._observers.values())

    def filter_observers(self, observers):
        last_counter = self._last_counter
        observers = (o() for o in observers)
        return [o for o in observers if o is not None and o.last_call <= last_counter]

    def add_observer(self, observer):
        if observer:
            self._last_counter = max(observer.last_call, self._last_counter)
            self._observers[id(observer)] = ref(observer)

    def clear_observers(self):
        self._observers = {}


def _cell_to_container(reactive):
    return [c.create_container(reactive) for c in reactive.__reactive_cells__.values()]


class _NOVAL:
    pass


def _state_from_dict(state, reactive, dict_):
    for c in reactive.__reactive_cells__.values():
        val = dict_.get(c.name, _NOVAL)
        if val is not _NOVAL:
            state[c.index]._value = val


def _state_as_dict(state, reactive):
    result = {}
    for c in reactive.__reactive_cells__.values():
        result[c.name] = state[c.index]._value
    return result


# __pragma__ ('skip')
class ReactiveState(list):
    """The state of a reactive objects. Contains all cell container."""

    def __init__(self, reactive):
        self[:] = _cell_to_container(reactive)

    def from_dict(self, reactive, dict_):
        _state_from_dict(self, reactive, dict_)

    def as_dict(self, reactive):
        return _state_as_dict(self, reactive)
# __pragma__ ('noskip')


"""?
__pragma__("js", "{}", '''

var ReactiveState_ = function(reactive) {
    this.push.apply(this, _cell_to_container(reactive));
    return this
}
ReactiveState_.prototype = Object.create(Array.prototype);
ReactiveState_.prototype.from_dict = function(reactive, dict_) {
  _state_from_dict(this, reactive, dict_);
}
ReactiveState_.prototype.as_dict = function(reactive) {
  return _state_as_dict(this, reactive);
}

export var ReactiveState = function(reactive) {
    return new ReactiveState_(reactive);
}

ReactiveState_.prototype.__class__ = ReactiveState;
deque.__name__ = 'ReactiveState';
deque.__bases__ = [list];
''')
?"""


class Container(Subject):
    """A cell container that stores a cell's value."""
    __slots__ = ("_value", )

    def __init__(self, value):
        super().__init__()
        self._value = value
        # see ccore.pyx Container.__dealloc__
        rcontext.old_values.pop(id(self), None)

    def set_value(self, value):
        """Sets the containers value."""
        old_value = self._value
        self._value = value
        # __pragma__ ("opov")
        if old_value != value or type(old_value) != type(value):
            # __pragma__ ("noopov")
            self.changed(old_value)
            return True
        return False

    def get_value(self):
        """Returns the container's value"""
        self.touch()
        return self._value

    def __repr__(self):
        return "<{0}({1:x}): {2!r}>".format(self.__class__.__name__, id(self),
                                            self._value)


class ResetContainer(Container):
    def __init__(self, value):
        super().__init__(value)
        self._reset_value = value

    def set_value(self, value):
        old_value = self._value
        self._value = value
        if old_value != value or type(old_value) != type(value):
            def reset():
                self._value = self._reset_value

            with rcontext.atomic:
                self.changed(self._reset_value)
                rcontext.shift_callback(reset)
            return True
        return False


class CellBase:
    """A descriptor to assign cell containers to objects."""
    __slots__ = ("name", "index")
    __container__ = Container  # factory

    def __init__(self):
        self.name = None
        self.index = -1

    def __init_cell__(self, name, index):
        self.name = name
        self.index = index

    def __get__(self, holder, owner):
        if holder is None:
            return self

        return holder.__reactive_state__[self.index].get_value()

    def __set__(self, holder, value):
        holder.__reactive_state__[self.index].set_value(value)

    def get_container(self, holder):
        return holder.__reactive_state__[self.index]

    def set_container(self, holder, value):
        # if we replace the container we have to keep the rules
        value.take(self.get_container(holder))
        holder.__reactive_state__[self.index] = value
# __pragma__ ('noecom')



import unittest
import larch.reactive as ra
import larch.reactive.rcollections as rc
from larch.reactive.pointer import _ExpressionOperandRoot
import pickle
# import larch.pickle as pickle
import gc
import weakref
import warnings
import logconfig
from contextlib import contextmanager
from operator import attrgetter


print("reactive.core", ra.core)


@contextmanager
def collect_warnings():
    messages = []

    def collect(message, category, filename, lineno, file=None, line=None):
        messages.append(message)

    old_warning = warnings.showwarning
    warnings.showwarning = collect
    try:
        yield messages
    finally:
        warnings.showwarning = old_warning


ra.add_pickle_extensions()


class RContextTester(ra.Reactive):
    a = ra.Cell(1)
    b = ra.Cell(2)
    c = ra.Cell(3)
    d = ra.Cell(4)
    calls = ra.MakeCell(list)

    @ra.rule(2)
    def _rule_b(self):
        if not ra.rcontext.inside_touch:
            raise RuntimeError("inside touch wrong")

        with ra.untouched():
            if ra.rcontext.inside_touch:
                raise RuntimeError("inside touch wrong")

            self.calls.append(("_rule_b in", self.a, self.b, self.c, self.d))

        self.c = self.b + 2
        self.d = self.b + 3

        with ra.untouched():
            self.calls.append(("_rule_b out", self.a, self.b, self.c, self.d))

    @ra.rule(3)
    def _rule_c(self):
        with ra.untouched():
            self.calls.append(("_rule_c in", self.a, self.b, self.c, self.d))

        self.d = self.c + 4

        with ra.untouched():
            self.calls.append(("_rule_c out", self.a, self.b, self.c, self.d))

    @ra.rule(4)
    def _rule_d(self):
        self.d
        with ra.untouched():
            self.calls.append(("_rule_d", self.a, self.b, self.c, self.d))

    @ra.rule(1)
    def _rule_a(self):
        with ra.untouched():
            self.calls.append(("_rule_a in", self.a, self.b, self.c, self.d))

        self.d = self.a + 3
        self.c = self.a + 2
        self.b = self.a + 1

        with ra.untouched():
            self.calls.append(("_rule_a out", self.a, self.b, self.c, self.d))


class RuleOrderTester(ra.Reactive):
    def __init__(self):
        self.called = []

    @ra.rule
    def _rule1(self):
        self.called.append("rule1")

    @ra.rule
    def _rule2(self):
        self.called.append("rule2")
        yield

    @ra.rule
    def _arule3(self):
        self.called.append("rule3")


class ErrorTester(ra.Reactive):
    callback_error = ra.Cell(0)
    recursive_error = ra.Cell(0)

    @ra.rule
    def _rule_callback_error(self):
        if self.callback_error:
            def outside():
                raise RuntimeError("callback_error")

            ra.rcontext.push_callback(outside)

    @ra.rule
    def _rule_recursive(self):
        if self.recursive_error:
            self.recursive_error += 1


class StopRuleTester(ra.Reactive):
    called = ra.Cell(0)
    iter_called = ra.Cell(0)

    @ra.rule
    def _rule_stop(self):
        if self.called == 1:
            # silently stop rule
            raise StopIteration()

    @ra.rule
    def _rule_iter_stop(self):
        called = self.iter_called
        yield
        if called == 1:
            # silently stop rule
            raise StopIteration()
        elif called == 2:
            raise RuntimeError()


class RContextTest(unittest.TestCase):
    def test_call_row(self):
        t = RContextTester()

        calls = [('_rule_a in', 1, 2, 3, 4),
                 ('_rule_a out', 1, 2, 3, 4),
                 ('_rule_b in', 1, 2, 3, 4),
                 ('_rule_b out', 1, 2, 4, 5),
                 ('_rule_c in', 1, 2, 4, 5),
                 ('_rule_c out', 1, 2, 4, 8),
                 ('_rule_d', 1, 2, 4, 8)]

        self.assertEqual(t.calls, calls)
        del t.calls[:]

        # tests also untouched
        t.a = 2
        calls = [('_rule_a in', 2, 2, 4, 8),
                 ('_rule_a out', 2, 3, 4, 5),
                 ('_rule_b in', 2, 3, 4, 5),
                 ('_rule_b out', 2, 3, 5, 6),
                 ('_rule_c in', 2, 3, 5, 6),
                 ('_rule_c out', 2, 3, 5, 9),
                 ('_rule_d', 2, 3, 5, 9)]

        self.assertEqual(t.calls, calls)
        del t.calls[:]

        t.b = 1
        calls = [('_rule_b in', 2, 1, 5, 9),
                 ('_rule_b out', 2, 1, 3, 4),
                 ('_rule_c in', 2, 1, 3, 4),
                 ('_rule_c out', 2, 1, 3, 7),
                 ('_rule_d', 2, 1, 3, 7)]

        self.assertEqual(t.calls, calls)
        del t.calls[:]

        t.c = 1
        calls = [('_rule_c in', 2, 1, 1, 7),
                 ('_rule_c out', 2, 1, 1, 5),
                 ('_rule_d', 2, 1, 1, 5)]

        self.assertEqual(t.calls, calls)

    def test_atomic(self):
        t = RContextTester()
        del t.calls[:]

        # from pudb import set_trace; set_trace()
        with ra.rcontext.atomic:
            t.a = 2
            t.d = 5

        calls = [('_rule_a in', 2, 2, 4, 5),
                 ('_rule_a out', 2, 3, 4, 5),
                 ('_rule_b in', 2, 3, 4, 5),
                 ('_rule_b out', 2, 3, 5, 6),
                 ('_rule_c in', 2, 3, 5, 6),
                 ('_rule_c out', 2, 3, 5, 9),
                 ('_rule_d', 2, 3, 5, 9)]

        self.assertEqual(t.a, 2)
        self.assertEqual(t.b, 3)
        self.assertEqual(t.c, 5)
        self.assertEqual(t.d, 9)

        self.assertEqual(t.calls, calls)
        del t.calls[:]

        with ra.rcontext.atomic:
            t.a = 1
            t.b = 8

        calls = [('_rule_a in', 1, 8, 5, 9),
                 ('_rule_a out', 1, 2, 3, 4),
                 ('_rule_b in', 1, 2, 3, 4),
                 ('_rule_b out', 1, 2, 4, 5),
                 ('_rule_c in', 1, 2, 4, 5),
                 ('_rule_c out', 1, 2, 4, 8),
                 ('_rule_d', 1, 2, 4, 8)]

        self.assertEqual(t.a, 1)
        self.assertEqual(t.b, 2)
        self.assertEqual(t.c, 4)
        self.assertEqual(t.d, 8)
        self.assertEqual(t.calls, calls)

    def test_silent(self):
        t = RContextTester()
        del t.calls[:]

        with ra.silent():
            t.a = 10

        self.assertEqual(t.calls, [])

    def test_misc(self):
        t = RContextTester()
        # should just not raise any exceptions
        ra.rcontext.rounds
        tc = ra.CellAgent(t)
        map(repr, tc.a.get_observers())
        self.assertIsInstance(tc.a, ra.Container)

    def test_callback_error(self):
        with collect_warnings() as messages:
            t = ErrorTester()
            t.callback_error = True

        self.assertEqual(len(messages), 1)
        self.assertIn("callback_error", repr(messages))

    def test_recursive_error(self):
        try:
            t = ErrorTester()
            t.recursive_error = 1
            self.assertFalse()
        except RuntimeError as e:
            self.assertIn("possible endless recursion", repr(e))


class RemoveWatcher(object):
    obj_removed = False

    def __init__(self, *args, **kwargs):
        super(RemoveWatcher, self).__init__(*args, **kwargs)
        self.__class__.obj_removed = False

    def __del__(self):
        self.__class__.obj_removed = True


@ra.reactive
class PointerChild(RemoveWatcher):
    a = ra.Cell(1)


@ra.reactive
class PointerChildChild(PointerChild):
    rule_called = False

    @ra.rule
    def _rule_is_reactive(self):
        self.rule_called = True


class PointerParent(RemoveWatcher, ra.Reactive):
    child = ra.Cell()
    children = ra.MakeCell(list)

    @ra.rule
    def _rule_set_children(self):
        self.children = [self.child]


class PointerTester:
    a = 1
    b = 2
    sum = ra.SELF.a + ra.SELF.b
    or_ = ra.SELF.a | ra.SELF.b
    and_ = ra.SELF.a & ra.SELF.b


class PointerObserver(RemoveWatcher, ra.Reactive):
    parent = ra.Cell()
    a = ra.SELF.parent.child.a
    old_a = 0

    @ra.rule
    def _rule_a(self):
        try:
            self.old_a = ra.OldAgent(self).a
        except ra.ResolveError:
            pass

        if ra.OldAgent(self).old_a != self.old_a:
            raise RuntimeError("not same")


def add(a, b):
    return 2 * a + b


class PointerTest(unittest.TestCase):
    def test_strong_ref(self):
        four = ra.Pointer(4)
        self.assertEqual(four(), 4)

    def test_delegator(self):
        c = PointerChild()
        p = PointerParent()
        o = PointerObserver()

        def wrong():
            o.a

        self.assertRaises(ra.ResolveError, wrong)

        o.parent = p
        self.assertRaises(ra.ResolveError, wrong)

        p.child = c
        self.assertEqual(o.a, 1)

        c.a = 10
        self.assertEqual(o.a, 10)
        self.assertEqual(o.old_a, 1)

        o.a = 20
        self.assertEqual(c.a, 20)
        self.assertEqual(o.old_a, 10)

        t = PointerTester()
        self.assertEqual(t.sum, 3)
        self.assertEqual(t.or_, 1)
        self.assertEqual(t.and_, 2)

    def test_same(self):
        o = PointerObserver(parent=PointerParent(child=PointerChild()))
        parent = o.parent

        pparent = ra.Pointer(parent)
        pchild = ra.Pointer(parent.child)

        p_pa = pparent.child.a
        p_pc = pparent.child

        self.assertEqual(p_pa, p_pa)
        self.assertEqual(p_pa, p_pc.a)
        self.assertEqual(p_pc.a, p_pa)
        self.assertEqual(p_pa, pchild.a)
        self.assertEqual(pchild.a, p_pa)
        self.assertNotEqual(p_pa, p_pc)
        self.assertNotEqual(p_pa, 2)

        p_pa1 = ra.Pointer(p_pa)
        self.assertEqual(p_pa1, p_pa)

    def test_single_pointer(self):
        c = PointerChild()
        pc = ra.Pointer(c)

        self.assertEqual(pc.a(), c.a)
        pc.a(2)
        self.assertEqual(c.a, 2)
        self.assertEqual(len(pc.a.__state__), 1)

        self.assertFalse(ra.Pointer())

    def test_pointer1(self):
        o = PointerObserver(parent=PointerParent(child=PointerChild()))

        pa = ra.Pointer(o).parent.child.a
        self.assertEqual(type(pa), ra.Pointer)
        self.assertEqual(str(pa), ".parent.child.a")
        self.assertEqual(hash(pa), hash(pa.__state__.path))

        o.a = 10
        self.assertEqual(o.parent.child.a, 10)
        self.assertEqual(pa(), 10)

    def test_pointer2(self):
        o = PointerObserver(parent=PointerParent(child=PointerChild()))

        pchild = ra.Pointer(o).parent.children[0]
        pa = ra.Pointer(o).parent.children[0].a

        pchild(None)
        self.assertEqual(bool(pa), False)
        self.assertEqual(o.parent.children, [None])

    def test_pointer_map(self):
        parent = PointerParent(child=PointerChild())
        pparent = ra.Pointer(parent)
        pmap = ra.PointerMap(child=pparent.child, a=pparent.child.a, parent=parent)

        self.assertEqual(pmap.get("child"), parent.child)
        self.assertEqual(pmap.get("a"), parent.child.a)
        self.assertEqual(pmap.get("parent"), parent)
        self.assertEqual(pmap.raw("child"), pparent.child)

        self.assertRaises(KeyError, pmap.get, "wrong")
        self.assertEqual(sorted(pmap.keys()), ['a', 'child', 'parent'])
        self.assertTrue("child" in pmap)

        pmap.set("a", 10)
        self.assertEqual(parent.child.a, 10)

        pmap.update(dict(a=20))
        self.assertEqual(pmap.get("a"), 20)

        self.assertTrue("a" in pmap)

        pmap.remove("child")
        self.assertRaises(KeyError, pmap.get, "child")

        pmap.set("value", 10)
        self.assertEqual(pmap.get("value"), 10)

    def test_compare_dumy_state(self):
        p1 = ra.Pointer()["name"].dir | "test"
        p2 = ra.Pointer()["name"].dir | "test"
        self.assertEqual(p1, p2)

    def get_proxies(self):
        self.t = PointerTester()
        return ra.Pointer(self.t).a, ra.Pointer(self.t).b

    def test_add(self):
        p1, p2 = self.get_proxies()
        p3 = p1 + p2
        self.assertEqual(p3(), 3)

        p1("ab")
        p2("cd")
        self.assertEqual(p3(), "abcd")

        p4 = p3[2]
        self.assertEqual(p4(), "c")

        p3 = "-" + p1
        self.assertEqual(p3(), "-ab")

    def test_sub(self):
        p1, p2 = self.get_proxies()
        p3 = p2 - p1
        self.assertEqual(p3(), 1)

        p3 = 10 - p2
        self.assertEqual(p3(), 8)

    def test_mul(self):
        p1, p2 = self.get_proxies()
        p3 = p2 * p1
        self.assertEqual(p3(), 2)

        p3 = 10 * p2
        self.assertEqual(p3(), 20)

    def test_div(self):
        p1, p2 = self.get_proxies()
        p1(4.0)
        p3 = p2 / p1
        self.assertEqual(p3(), 0.5)

        p3 = 10 / p2
        self.assertEqual(p3(), 5)

    def test_and(self):
        p1, p2 = self.get_proxies()
        p3 = p2 & p1
        self.assertEqual(p3(), 1)

        p3 = 10 & p2
        self.assertEqual(p3(), 2)

        pu = ra.Pointer().a
        p3 = p2 & pu
        self.assertFalse(p3())

        p3 = False & p2
        self.assertFalse(p3())

    def test_or(self):
        p1, p2 = self.get_proxies()
        p3 = p2 | p1
        self.assertEqual(p3(), 2)

        p3 = 10 | p2
        self.assertEqual(p3(), 10)

        p1(0)
        p3 = False | p1
        self.assertEqual(p3(), 0)

        pu = ra.Pointer().a
        p3 = False | pu | p1
        self.assertEqual(p3(), 0)

    def test_call(self):
        p1, p2 = self.get_proxies()
        p3 = ra.PointerExpression.call(add, p1, b=p2)
        self.assertEqual(p3(), 4)

    def test_compare(self):
        p1 = ra.Pointer().attr1
        p2 = ra.Pointer().attr2
        p3 = p1 & p2

        p4 = pickle.loads(pickle.dumps(p3))
        self.assertEqual(p3, p4)
        self.assertNotEqual(p3, p2)

        p4 = p1 | p2
        self.assertNotEqual(p3, p4)

    def test_apply(self):
        p1, p2 = self.get_proxies()
        p3 = ra.PointerExpression.apply(tuple, p1, p2)
        self.assertEqual(p3(), (1, 2))

    def test_set(self):
        p1, p2 = self.get_proxies()
        p3 = p2 + p1
        self.assertRaises(ValueError, p3, 3)

        p1(2)
        self.assertEqual(self.t.a, 2)

        p = ra.Pointer(self.t)
        self.assertRaises(TypeError, p, 2)

    def test_pickle(self):
        p1, p2 = self.get_proxies()

        x = pickle.dumps((self.t, p1))
        t, y = pickle.loads(x)
        self.assertEqual(y(), 1)

        p3 = p2 + p1
        x = pickle.dumps((self.t, p3))
        t, y = pickle.loads(x)
        self.assertEqual(y(), 3)

        p3 = p2 | p1
        x = pickle.dumps((self.t, p3))
        t, y = pickle.loads(x)
        self.assertEqual(y(), 2)

        p3 = ra.Pointer().a + ra.Pointer().b
        x = pickle.dumps(p3)
        y = pickle.loads(x)
        p4 = ra.merge_pointers(ra.Pointer(self.t), y)
        self.assertEqual(p4(), 3)

        p1, p2 = self.get_proxies()
        p3 = ra.PointerExpression.call(add, p1, b=p2)
        x = pickle.dumps((p3, self.t))
        y, t = pickle.loads(x)
        self.assertEqual(y(), 4)

        p1, p2 = self.get_proxies()
        p3 = ra.PointerExpression.call(add, p1, b=3)
        x = pickle.dumps(p3)
        y = pickle.loads(x)
        p4 = ra.merge_pointers(ra.Pointer(self.t), y)
        self.assertEqual(p4(), 5)
        self.assertEqual(y.__state__.delegate_get(self.t), 5)

    def test_backward_compatibility(self):
        ep = _ExpressionOperandRoot(add, [1, 2], {"a": 1})
        ep.__setstate__((add, [2, 3]))  # no exception
        self.assertEqual(ep.args, [2, 3])
        self.assertEqual(ep.kwargs, {})

    def test_equal_state(self):
        p = ra.Pointer()

        pin = p.a
        pout = pickle.loads(pickle.dumps(pin))
        self.assertEqual(pin.__state__, pout.__state__)

        pin = p.a + p.b
        pout = pickle.loads(pickle.dumps(pin))
        self.assertEqual(pin.__state__, pout.__state__)

        self.assertNotEqual((p.a + p.b).__state__, (p.a + p.c).__state__)
        self.assertNotEqual((p.a + p.b).__state__, (p.a - p.b).__state__)
        self.assertNotEqual((p.a | p.b).__state__, (p.a | p.c).__state__)
        self.assertNotEqual((p.a | p.b).__state__, (p.a + p.c).__state__)
        self.assertNotEqual(p.a.__state__, (p.a | p.c).__state__)

    def test_merge(self):
        # merge PointerState to PointerState
        p1 = ra.Pointer().a.b
        p2 = ra.Pointer().c.d
        p3 = ra.merge_pointers(p1, p2)
        self.assertEqual(str(p3), ".a.b.c.d")

        # merge PointerState to PointerExpression

        # _ExpressionOperandRoot.merge_to
        p1 = ra.Pointer().a
        p2 = ra.Pointer().b
        p3 = p1 + p2
        self.assertRaises(ra.ResolveError, p3)

        t = PointerTester()
        p4 = ra.merge_pointers(ra.Pointer(t), p3)
        self.assertEqual(p4(), 3)

        # _ExpressionRoot.merge_to
        p5 = p1 | p2
        self.assertFalse(p5())
        p6 = ra.merge_pointers(ra.Pointer(t), p5)
        self.assertEqual(p6(), 1)

        # merge PointerExpression to PointerState
        t.a = [1]
        t.b = [2]
        p5 = ra.merge_pointers(p4, ra.Pointer()[1])
        self.assertEqual(p5(), 2)

    def test_nothing(self):
        self.assertFalse(ra.NOTHING)

    def test_bool(self):
        p = PointerParent()
        pchild = ra.Pointer(p).child
        self.assertTrue(bool(pchild))

        pchild = ra.Pointer().child
        self.assertFalse(bool(pchild))

    def test_backward_load(self):
        data = [
            ("<Pointer-<class 'larch.reactive.pointer.NOTHING'>.test>",
             (b"\x80\x04\x95'\x00\x00\x00\x00\x00\x00\x00\x84\xc1\x8f"
              b"\x19-\x84\xf1\xab,R)\x81\x94\x84\xf7K\xb4p\x8c\x04test"
              b"\x94\x85\x94\x8c\x01a\x94\x87\x94b\x85\x94R\x94.")),
            ("<Pointer-and(<class 'larch.reactive.pointer.NOTHING'>.b,)>",
             (b'\x80\x04\x95>\x00\x00\x00\x00\x00\x00\x00\x84\xc1\x8f'
              b'\x19-\x84\xc3\xa0QH)\x81\x94\x84\xaf\x85\xd6})\x81\x94'
              b'\x84\xf1\xab,R)\x81\x94\x84\xf7K\xb4p\x8c\x01b\x94\x85'
              b'\x94\x8c\x01a\x94\x87\x94b\x85\x94b)\x8c\x00\x94\x87\x94b\x85\x94R\x94.')),
            ("<Pointer-or(<class 'larch.reactive.pointer.NOTHING'>.b,)>",
             (b'\x80\x04\x95>\x00\x00\x00\x00\x00\x00\x00\x84\xc1\x8f'
              b'\x19-\x84\xc3\xa0QH)\x81\x94\x84\xc9V-\x08)\x81\x94\x84'
              b'\xf1\xab,R)\x81\x94\x84\xf7K\xb4p\x8c\x01b\x94\x85\x94'
              b'\x8c\x01a\x94\x87\x94b\x85\x94b)\x8c\x00\x94\x87\x94b\x85\x94R\x94.')),
            (("<Pointer-add(<class 'larch.reactive.pointer.NOTHING'>.a, "
              "<class 'larch.reactive.pointer.NOTHING'>.b){}>"),
             (b'\x80\x04\x95p\x00\x00\x00\x00\x00\x00\x00\x84'
              b'\xc1\x8f\x19-\x84\xc3\xa0QH)\x81\x94\x84\x9c['
              b'\x91K)\x81\x94\x8c\rtest_reactive\x94\x8c\x03add'
              b'\x94\x93\x94\x84\xf1\xab,R)\x81\x94\x84\xf7K\xb4p'
              b'\x8c\x01a\x94\x85\x94h\x06\x87\x94b\x84\xf1\xab,R)'
              b'\x81\x94\x84\xf7K\xb4p\x8c\x01b\x94\x85\x94h\x06\x87'
              b'\x94b\x86\x94}\x94\x87\x94b)\x8c\x00\x94\x87\x94b\x85\x94R\x94.')),
            (("<Pointer-sum(<class 'larch.reactive.pointer.NOTHING'>.a, "
              "<class 'larch.reactive.pointer.NOTHING'>.b){}>"),
             (b'\x80\x04\x95k\x00\x00\x00\x00\x00\x00\x00\x84\xc1\x8f\x19-\x84\xc3\xa0QH)'
              b'\x81\x94\x84\x0e\xb4\xa4:)\x81\x94\x8c\x08builtins\x94\x8c\x03sum\x94\x93'
              b'\x94\x84\xf1\xab,R)\x81\x94\x84\xf7K\xb4p\x8c\x01a\x94\x85\x94h\x06\x87'
              b'\x94b\x84\xf1\xab,R)\x81\x94\x84\xf7K\xb4p\x8c\x01b\x94\x85\x94h\x06\x87'
              b'\x94b\x86\x94}\x94\x87\x94b)\x8c\x00\x94\x87\x94b\x85\x94R\x94.'))]

        for src, pkl in data:
            r = pickle.loads(pkl)
            self.assertEqual(repr(r), src)


class AgentTester(ra.Reactive):
    a = ra.Cell(1)
    old_a = 1
    old_ap = 1
    to_old_a = ra.SELF.old_a

    @ra.rule
    def _rule_a(self):
        self.old_a = ra.old(self).a
        self.old_ap = ra.old(self.a)


class AgentTest(unittest.TestCase):

    def test_old_agent(self):
        t = AgentTester()

        t.a = 10
        self.assertEqual(t.old_a, 1)

        t.a = 20
        self.assertEqual(t.old_a, 10)

        t2 = AgentTester(a=ra.Pointer(t).a)
        self.assertEqual(t2.old_ap, 20)

        t.a = 30
        self.assertEqual(t2.old_ap, 20)

        t.a = 40
        self.assertEqual(t2.old_ap, 30)

    def test_cell_agent(self):
        t1 = AgentTester()
        t2 = AgentTester()

        t1_cells = ra.CellAgent(t1)
        t2_cells = ra.CellAgent(t2)

        self.assertEqual(
            t1_cells.a.get_observers()[0].method,
            AgentTester._rule_a.__rule__[-1])
        self.assertEqual(t1_cells.a.get_observers()[0].holder(), t1)

        t1_cells.a = t2_cells.a

        t1.a = 10
        self.assertEqual(t2.a, 10)
        self.assertEqual(t1.old_a, 1)
        self.assertEqual(t2.old_a, 1)

        t2.a = 20
        self.assertEqual(t1.a, 20)
        self.assertEqual(t1.old_a, 10)
        self.assertEqual(t2.old_a, 10)

        def wrong():
            t1_cells.a = 1

        self.assertRaises(TypeError, wrong)

    def test_cell_agent_pointer(self):
        t1 = AgentTester()
        t2 = AgentTester()

        pa1 = ra.Pointer(t1).a
        pa2 = ra.Pointer(t2).a

        self.assertEqual(
            ra.cell(pa1).get_observers()[0].method,
            AgentTester._rule_a.__rule__[-1])
        self.assertEqual(ra.cell(pa1).get_observers()[0].holder(), t1)

        ra.cell(pa1, ra.cell(pa2))

        pa1(10)
        self.assertEqual(t2.a, 10)
        self.assertEqual(t1.old_a, 1)
        self.assertEqual(t2.old_a, 1)

        pa2(20)
        self.assertEqual(t1.a, 20)
        self.assertEqual(t1.old_a, 10)
        self.assertEqual(t2.old_a, 10)

        self.assertEqual(len(pa2), 0)
        self.assertRaises(TypeError, iter, pa2)

        del t1
        gc.collect()
        self.assertRaises(ra.ResolveError, pa1, 1)

    def test_errors(self):
        t = AgentTester()
        self.assertRaises(TypeError, ra.cell, ra.Pointer(t).a, 2)

        self.assertEqual(ra.old(ra.Pointer(t).old_a), 1)

        self.assertRaises(ra.ResolveError, ra.cell, ra.Pointer().a)

        def get_non_cell():
            ra.cell(t).to_old_a

        self.assertRaises(TypeError, get_non_cell)


class RobustString(str):

    def convert(self, other):
        return str(other)


class CellTester(ra.Reactive):
    a = ra.Cell()
    b = ra.TypeCell("")
    c = ra.ResetCell(False)
    d = ra.MakeCell(list, (1, 2, 3))
    e = ra.MakeTypeCell(list, (1, 2, 3))
    f = ra.TypeCell(RobustString(""))
    g = ra.TypeCell("", repr)

    @ra.rule
    def _rule_check_observer_callback(self):
        self.c


class NoString(object):

    def __str__(self):
        raise RuntimeError("no string")

    def __repr__(self):
        return "<NoString>"


class CellTest(unittest.TestCase):

    def test_cells(self):
        t = CellTester(a=1, b=2, c=True)

        self.assertEqual(t.a, 1)
        self.assertEqual(t.b, "2")
        self.assertEqual(t.c, False)
        self.assertEqual(t.d, [1, 2, 3])
        self.assertEqual(t.e, [1, 2, 3])

        t.d = (4, 5, 6)
        t.e = (4, 5, 6)

        self.assertEqual(t.d, (4, 5, 6))
        self.assertEqual(t.e, [4, 5, 6])

        t.c = True
        self.assertEqual(t.c, False)

        t.f = 10
        self.assertEqual(t.f, "10")

        try:
            t.b = NoString()
            self.assertFalse(True)
        except RuntimeError as e:
            self.assertEqual(e.args[1], CellTester.b)

        t.g = NoString()
        self.assertEqual(t.g, repr(NoString()))


class MultiBaseA(ra.Reactive):
    a = ra.Cell(1)
    b = ra.MakeCell(int, 2)
    calls = ra.MakeCell(list)

    @ra.rule(2)
    def _rule_a(self):
        with ra.rcontext.untouched:
            calls = self.calls

        calls.append(("a", self.a, self.b))


class MultiBaseB(ra.Reactive):
    b = ra.Cell(3)
    c = ra.Cell(4)
    d = ra.Cell(5)
    calls = ra.MakeCell(list)

    @ra.rule(3)
    def _rule_b(self):
        with ra.rcontext.untouched:
            calls = self.calls

        calls.append(("b", self.b, self.c, self.d))

    @ra.rule(4)
    def _rule_c(self):
        with ra.rcontext.untouched:
            calls = self.calls

        calls.append(("c", self.b))


class MulitSubC(MultiBaseA, MultiBaseB):
    a = ra.Cell(6)
    d = ra.Cell(7)
    e = ra.Cell(8)
    calls = ra.MakeCell(list)

    @ra.rule(1)
    def _rule_c(self):
        with ra.rcontext.untouched:
            calls = self.calls

        calls.append(("c", self.b, self.c))

    @ra.rule(5)
    def _rule_d(self):
        with ra.rcontext.untouched:
            calls = self.calls

        calls.append(("d", self.a, self.b, self.c, self.d, self.e))


class Observer(ra.Reactive):
    subject = ra.Cell()
    subjects_value = -1

    @ra.rule
    def _rule_set_value(self):
        if self.subject:
            self.subjects_value = self.subject.value


class Subject(ra.Reactive):
    value = ra.Cell(0)


class RuleOrder(ra.Reactive):
    def __init__(self):
        self.call_order = []

    @ra.rule
    def rule9(self):
        self.call_order.append(9)

    @ra.rule
    def rule5(self):
        self.call_order.append(5)

    @ra.rule
    def rule0(self):
        self.call_order.append(0)


class CircRef(ra.Reactive):
    value = ra.Cell()


class RuleBase(ra.Reactive):
    value = ra.Cell()

    def __init__(self):
        self.call_order = []

    @ra.rule
    def rule0(self):
        value = self.value
        yield
        self.call_order.append((0, value))


class RuleSuper(RuleBase):
    @ra.rule
    def rule0(self):
        # calling super generator rules
        g = super(RuleSuper, self).rule0()
        next(g)
        yield
        all(g)
        self.call_order.append((1, self.value))


class TestReactive(unittest.TestCase):

    def test_pickle(self):
        o1 = PointerObserver(parent=PointerParent(child=PointerChild()))

        o1.parent.child.a = 10
        self.assertEqual(o1.old_a, 1)

        o1.parent.child.a = 20
        self.assertEqual(o1.old_a, 10)

        # from pudb import set_trace; set_trace()
        picklestr = pickle.dumps(o1)

        o2 = pickle.loads(picklestr)
        self.assertEqual(o2.a, 20)
        self.assertEqual(o2.parent.child.a, 20)
        self.assertEqual(o2.old_a, 10)

    def test_multi(self):
        c = MulitSubC()

        self.assertEqual(c.a, 6)
        self.assertEqual(c.b, 2)
        self.assertEqual(c.c, 4)
        self.assertEqual(c.d, 7)
        self.assertEqual(c.e, 8)

        calls = [('c', 2, 4),
                 ('a', 6, 2),
                 ('b', 2, 4, 7),
                 ('d', 6, 2, 4, 7, 8)]

        self.assertEqual(c.calls, calls)
        # print ra.dump(c)

    def test_cells_of(self):
        self.assertEqual(list(map(attrgetter("name"), ra.cells_of(MulitSubC))),
                         ['b', 'c', 'a', 'd', 'e', 'calls'])

        self.assertEqual(
            list(map(attrgetter("name"), ra.cells_of(MulitSubC()))),
            ['b', 'c', 'a', 'd', 'e', 'calls'])

    def test_weak_refs(self):
        o = PointerObserver(parent=PointerParent(child=PointerChild()))
        self.assertEqual(PointerObserver.obj_removed, False)
        self.assertEqual(PointerParent.obj_removed, False)
        self.assertEqual(PointerChild.obj_removed, False)
        self.assertEqual(o.a, 1)

        o.parent.child = None
        gc.collect()
        self.assertEqual(PointerChild.obj_removed, True)
        observer = ra.cell(o).parent.get_observers()[0]

        del o
        gc.collect()
        self.assertEqual(PointerObserver.obj_removed, True)
        self.assertEqual(PointerParent.obj_removed, True)
        self.assertEqual(repr(observer)[:23], "<Rule obsolete _rule_a ")

    def test_decouple(self):
        # from pudb import set_trace; set_trace()
        subject1 = Subject()
        subject2 = Subject()
        observer = Observer(subject=subject1)

        self.assertEqual(observer.subjects_value, 0)

        subject1.value = 1
        self.assertEqual(observer.subjects_value, 1)

        subject2.value = 2
        self.assertEqual(observer.subjects_value, 1)

        observer.subject = subject2
        self.assertEqual(observer.subjects_value, 2)

        subject1.value = 10
        self.assertEqual(observer.subjects_value, 2)

        subject2.value = 3
        self.assertEqual(observer.subjects_value, 3)

    def test_rule_order(self):
        ro = RuleOrder()
        self.assertEqual(ro.call_order, [9, 5, 0])

    def test_subclass(self):
        ccp = PointerChildChild()
        self.assertEqual(ccp.rule_called, True)

    def test_circula_ref(self):
        gc.disable()

        a = CircRef()
        b = CircRef()
        # circular reference
        a.value = b
        b.value = a

        c = weakref.ref(a)
        del a
        del b
        # not garage collected (circular refcount)
        self.assertTrue(bool(c()))
        gc.collect()
        gc.enable()
        self.assertFalse(bool(c()))

    def test_rule_super(self):
        ro = RuleSuper()
        self.assertEqual(ro.call_order, [(0, None), (1, None)])
        ro.call_order = []
        ro.value = 5
        self.assertEqual(ro.call_order, [(0, 5), (1, 5)])

    def test_stop_rule(self):
        r = StopRuleTester()

        with collect_warnings() as messages:
            r.iter_called = 1
        self.assertFalse(messages)

        with collect_warnings() as messages:
            r.called = 1
        self.assertFalse(messages)

        # nothing happens
        logconfig.logstream.truncate(0)
        with collect_warnings() as messages:
            r.iter_called = 2
        self.assertIn("RuntimeError", str(messages[0]))


class MiscTester(ra.Reactive):
    a = ra.ResetCell(False)
    b = ra.ResetCell(False)
    will_raise = ra.ResetCell(False)
    will_raise_outside = ra.ResetCell(False)
    c = ra.ResetCell(False)
    d = ra.ResetCell(False)
    double_count = 0

    e = ra.ResetCell(False)
    f = ra.ResetCell(False)
    g = ra.ResetCell(False)
    reinsert_count = 0

    @ra.rule
    @ra.untouched()
    def _rule_b(self):
        if not ra.rcontext.inside_rule:
            raise RuntimeError("inside rule wrong")

        with ra.touched():
            self.b

    @ra.rule
    def _rule_will_raise(self):
        if self.will_raise:
            raise RuntimeError("raise exception")

    @ra.rule
    def _rule_double(self):
        self.c
        self.d
        self.double_count += 1

    @ra.rule
    def _rule_reinsert(self):
        self.reinsert_count += 1
        if self.e and not self.f:
            self.f = True

    @ra.rule
    def _rule_raise_outside(self):
        def raise_outside():
            raise RuntimeError("")

        if self.will_raise_outside:
            ra.call_outside(raise_outside)

    @ra.rule
    def _rule_coroutine(self):
        if self.g:
            self.co_call = [(0, ra.rcontext.transaction_level)]
            yield
            self.co_call.append((1, ra.rcontext.transaction_level))
            yield
            self.co_call.append((2, ra.rcontext.transaction_level))


class Control(ra.Reactive):
    view = ra.Cell()

    def __init__(self, view):
        self.view = view
        self.calls = []

    @ra.rule
    def _rule_value_changed(self):
        self.calls.append(self.view.value)


class View(ra.Reactive):
    """A situation that failed in practice"""
    value = ra.Cell()
    old_value = None

    @ra.rule(1)
    def _rule_create_control(self):
        if self.old_value is None:
            self.control = Control(self)

        self.old_value = self.value


class DirectRuleCall(ra.Reactive):
    a = ra.Cell(1)
    called = None

    @ra.rule
    def _rule_activation(self, activate=None):
        if activate is None:
            self.called = self.a

        elif activate is False:
            self.called = None

        else:
            self.a
            self.called = None


class MissingObserverTest(ra.Reactive):
    cell1 = ra.Cell(False)
    cell2 = ra.Cell(0)

    @ra.atomic()
    def start_test(self):
        self.cell2 = 1
        self.cell1 = True

    @ra.rule
    def _rule_for_cell2(self):
        self.value2 = self.cell2  # touch

    @ra.rule
    def _rule_missing(self):
        if self.cell1:
            self.value = self.cell2

    @ra.rule
    def _rule_fire_cell2(self):
        if self.cell1:
            self.cell2 = 2


class MiscTest(unittest.TestCase):
    """to complete code coverage"""

    def test_callback_stack(self):
        called_ids = []

        def call_inner():
            called_ids.append("inner")

        def call_outer():
            with ra.atomic():
                called_ids.append("inner")
            self.assertEqual(called_ids, ["inner"])
            called_ids.append("outer")

        with ra.atomic():
            ra.call_outside(call_outer)
            ra.call_outside(call_inner)

        self.assertEqual(called_ids, ["inner", "outer", "inner"])

    def test_missing_observer(self):
        # tests a bug
        m = MissingObserverTest()
        m.start_test()
        self.assertEqual(m.value, 2)

    def test_doubled_rule(self):
        t = MiscTester()

        # without transaction
        t.double_count = 0
        t.c = 1
        t.d = 2
        self.assertEqual(t.double_count, 2)

        def check_after():
            self.assertEqual(t.double_count, 1)
            t.check_called = True

        with ra.atomic():
            self.assertFalse(ra.rcontext.inside_rule)
            t.c = 1
            t.c = 1
            t.d = 2

            observers = ra.rcontext.observers
            self.assertEqual(len(observers), 2)
            self.assertEqual(observers[0], observers[1])
            t.double_count = 0
            ra.call_outside(check_after)

        self.assertEqual(t.double_count, 1)
        self.assertEqual(t.check_called, True)

        t.check_called = False
        ra.call_outside(check_after)
        self.assertEqual(t.check_called, True)

    def test_reinsert_rule(self):
        t = MiscTester()
        t.reinsert_count = 0
        t.e = True
        self.assertEqual(t.reinsert_count, 2)

    def test_empty_call_back(self):
        t = MiscTester()
        with ra.rcontext.atomic:
            t.a = True

        self.assertEqual(t.a, False)

    def test_dumy_call_back(self):
        t = MiscTester()
        with ra.rcontext.silent:
            t.b = True

        self.assertEqual(t.b, False)

    def test_rule_exception(self):
        t = MiscTester()

        logconfig.logstream.truncate(0)
        with collect_warnings() as messages:
            t.will_raise = True

        self.assertEqual(len(messages), 1)
        self.assertIn("_rule_will_raise", str(messages[0]))
        self.assertIn("RuntimeError: raise exception",
                      logconfig.logstream.getvalue())

    def test_outside_exception(self):
        t = MiscTester()

        with collect_warnings() as messages:
            t.will_raise_outside = True

        self.assertEqual(len(messages), 1)
        self.assertIn("raise_outside", str(messages[0]))

    def test_coroutine_rule(self):
        t = MiscTester()
        t.g = True
        self.assertEqual(t.co_call, [(0, 1), (1, 0), (2, 0)])

    def test_counter_error(self):
        # the Control _rule_value_changed was detached because
        # _rule_create_control has a smaller counter than _rule_value_changed
        view = View()
        view.value = 1
        self.assertEqual(view.control.calls, [1])
        view.value = 2
        self.assertEqual(view.control.calls, [1, 2])

    def test_init_rule_order(self):
        t = RuleOrderTester()
        self.assertEqual(t.called, ['rule1', 'rule2', 'rule3'])

    def test_direct_rule_call(self):
        o = DirectRuleCall()

        # rule is activated by default
        self.assertEqual(o.called, 1)

        o.a = 2
        self.assertEqual(o.called, 2)

        # rule is deactivated
        # from pudb import set_trace; set_trace()
        o._rule_activation(False)
        self.assertEqual(o.called, None)

        o.a = 3
        self.assertEqual(o.called, None)

        # rule reactivated
        o._rule_activation(True)
        self.assertEqual(o.called, None)

        o.a = 4
        self.assertEqual(o.called, 4)

    def test_agent_dir(self):
        t = RContextTester()
        tc = ra.CellAgent(t)
        self.assertEqual(dir(tc), ['a', 'b', 'c', 'calls', 'd'])


class ListTester(ra.Reactive):
    li = ra.MakeCell(rc.List)
    add = ra.ResetCell(None)

    def __init__(self):
        self.calls = []

    @ra.rule
    def _rule_add(self):
        if self.add is not None:
            self.li.append(self.add)

    @ra.rule
    def _rule_start(self):
        if self.li.__action_start__:
            self.calls.append(("start", self.li.__action_start__))

    @ra.rule
    def _rule_end(self):
        if self.li.__action_end__:
            self.calls.append(("end", self.li.__action_end__))


class ListTest(unittest.TestCase):
    def test_sandwich(self):
        lt = ListTester()
        lt.add = 1
        self.assertEqual(lt.calls, [('start', ('insert', (0, 1))),
                                    ('end', ('insert', (0, 1)))])
        self.assertEqual(lt.li, [1])

    def test_manipulation(self):
        lt = ListTester()
        lt.li[:] = [1, 2, 3, 4]

        calls = [('start', ('change', (slice(0, 0, None), [1, 2, 3, 4]))),
                 ('end', ('change', (slice(0, 0, None), [1, 2, 3, 4])))]

        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('change', (slice(2, 4, None), [5, 6, 7]))),
                 ('end', ('change', (slice(2, 4, None), [5, 6, 7])))]
        lt.li[-2:] = [5, 6, 7]
        self.assertEqual(lt.li, [1, 2, 5, 6, 7])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('delete', slice(3, 4, None))),
                 ('end', ('delete', slice(3, 4, None)))]

        del lt.li[-2]
        self.assertEqual(lt.li, [1, 2, 5, 7])
        self.assertEqual(lt.calls, calls)

        del lt.li[len(lt.li):]  # nothing changed
        self.assertEqual(lt.li, [1, 2, 5, 7])
        self.assertEqual(lt.calls, calls)

        lt.li[1:1] = []  # nothing changed
        self.assertEqual(lt.li, [1, 2, 5, 7])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('delete', slice(1, 2, None))),
                 ('end', ('delete', slice(1, 2, None)))]

        lt.li.remove(2)
        self.assertEqual(lt.li, [1, 5, 7])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('delete', slice(2, 3, None))),
                 ('end', ('delete', slice(2, 3, None)))]

        lt.li.pop(-1)
        self.assertEqual(lt.li, [1, 5])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('change', (slice(1, 2, None), [6]))),
                 ('end', ('change', (slice(1, 2, None), [6])))]

        lt.li[-1] = 6
        self.assertEqual(lt.li, [1, 6])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('delete', slice(0, 2, None))),
                 ('end', ('delete', slice(0, 2, None)))]

        lt.li[:] = []
        self.assertEqual(lt.li, [])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('insert', (0, 1))), ('end', ('insert', (0, 1)))]
        lt.li.insert(0, 1)
        self.assertEqual(lt.li, [1])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('extend', [2, 3])), ('end', ('extend', [2, 3]))]
        lt.li.extend([2, 3])
        self.assertEqual(lt.li, [1, 2, 3])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('extend', [4, 5])), ('end', ('extend', [4, 5]))]
        lt.li += [4, 5]
        self.assertEqual(lt.li, [1, 2, 3, 4, 5])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('imul', 2)), ('end', ('imul', 2))]
        lt.li *= 2
        self.assertEqual(lt.li, [1, 2, 3, 4, 5] * 2)
        self.assertEqual(lt.calls, calls)

        lt.li[:] = [1, 2, 3]
        lt.calls = []
        calls = [('start', ('order', 'reverse')),
                 ('end', ('order', 'reverse'))]
        lt.li.reverse()
        self.assertEqual(lt.li, [3, 2, 1])
        self.assertEqual(lt.calls, calls)

        lt.calls = []
        calls = [('start', ('order', 'sort')), ('end', ('order', 'sort'))]
        lt.li.sort()
        self.assertEqual(lt.li, [1, 2, 3])
        self.assertEqual(lt.calls, calls)

    def test_pickle(self):
        lt = ListTester()
        lt.li[:] = [1, 2, 3, 4]

        x = pickle.dumps(lt.li, -1)
        y = pickle.loads(x)

        self.assertEqual(y, lt.li)
        self.assertEqual(type(y), type(lt.li))


class DictTester(ra.Reactive):
    d = ra.MakeCell(rc.Dict)
    change = ra.ResetCell(None)

    def __init__(self):
        self.calls = []

    @ra.rule
    def _rule_change(self):
        if self.change is None:
            return

        action, key, value = self.change
        if action == "set":
            self.d[key] = value
        else:
            del self.d[key]

    @ra.rule
    def _rule_start(self):
        if self.d.__action_start__:
            self.calls.append(("start", self.d.__action_start__))

    @ra.rule
    def _rule_end(self):
        if self.d.__action_end__:
            self.calls.append(("end", self.d.__action_end__))


class DictTest(unittest.TestCase):

    def test_sandwich(self):
        dt = DictTester()
        # from pudb import set_trace; set_trace()
        dt.change = ("set", "test", 1)
        self.assertEqual(dt.calls, [('start', ('change', ('test', 1))),
                                    ('end', ('change', ('test', 1)))])
        self.assertEqual(dt.d, {'test': 1})

    def test_manipulation(self):
        dt = DictTester()
        dt.d[1] = 2

        calls = [('start', ('change', (1, 2))),
                 ('end', ('change', (1, 2)))]

        self.assertEqual(dt.calls, calls)

        # delete
        dt.calls = []
        del dt.d[1]
        calls = [('start', ('delete', 1)), ('end', ('delete', 1))]
        self.assertEqual(dt.calls, calls)

        # update
        dt.calls = []
        dt.d.update({1: 2, 2: 3, 3: 4})

        calls = [('start', ('update', (({1: 2, 2: 3, 3: 4},), {}))),
                 ('end', ('update', (({1: 2, 2: 3, 3: 4},), {})))]
        self.assertEqual(dt.calls, calls)

        # pop
        dt.calls = []
        r = dt.d.pop(1)
        calls = [('start', ('delete', 1)),
                 ('end', ('delete', 1))]
        self.assertEqual(r, 2)
        self.assertEqual(dt.calls, calls)

        dt.calls = []
        r = dt.d.pop(None, 1)
        self.assertEqual(r, 1)
        self.assertEqual(dt.calls, [])

        dt.calls = []
        self.assertRaises(KeyError, dt.d.pop, None)
        self.assertEqual(dt.calls, [])

        # delete unknown key
        dt.calls = []
        self.assertRaises(KeyError, dt.d.__delitem__, None)
        self.assertEqual(dt.calls, [])  # nothing changed

        # setdefault
        dt.calls = []
        calls = [('start', ('change', ('default', 1))),
                 ('end', ('change', ('default', 1)))]

        dt.d.setdefault("default", 1)
        self.assertEqual(dt.calls, calls)

        # popitem
        dt.calls = []
        item = dt.d.popitem()
        calls = [('start', ('delete', item[0])), ('end', ('delete', item[0]))]
        self.assertEqual(dt.calls, calls)

        # clear
        dt.calls = []
        calls = [('start', ('clear', None)), ('end', ('clear', None))]
        dt.d.clear()
        self.assertEqual(dt.calls, calls)
        self.assertRaises(KeyError, dt.d.popitem)

    def test_pickle(self):
        dt = DictTester()
        dt.d.update({1: 2, 2: 3, 3: 4})

        ds = pickle.dumps(dt.d, -1)
        d = pickle.loads(ds)

        self.assertEqual(d, dt.d)
        self.assertEqual(type(d), type(dt.d))


class TempConverter(ra.Reactive):
    F = ra.Cell(32)
    C = ra.Cell(0)

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @ra.rule
    def convert_fahrenheit(self):
        self.F = self.C * 1.8 + 32

    @ra.rule
    def convert_celsius(self):
        self.C = (self.F - 32) / 1.8


class DocTest(unittest.TestCase):
    def test_temp(self):
        # from pudb import set_trace; set_trace()
        tc = TempConverter(C=100)
        self.assertEqual(tc.C, 100)
        self.assertEqual(tc.F, 212)

        tc.F = 32.0
        self.assertEqual(tc.C, 0)
        self.assertEqual(tc.F, 32)

        tc.C = -40.0
        self.assertEqual(tc.C, -40)
        self.assertEqual(tc.F, -40)


if __name__ == "__main__":
    unittest.main()



# cython: language_level = 3
# distutils: language = c++
import types
import sys
import warnings
import logging
from functools import update_wrapper
from cpython cimport (
    Py_INCREF, Py_DECREF, PyThread_get_thread_ident, PyList_GET_ITEM,
    PyObject_GetAttrString, PyBytes_AS_STRING, PyObject)
from cpython.tuple cimport PyTuple_GET_ITEM, PyTuple_SET_ITEM, PyTuple_New
from cpython.weakref cimport PyWeakref_GET_OBJECT, PyWeakref_NewRef
from cython.operator cimport dereference as deref, preincrement as inc
from libcpp.deque cimport deque
from ._handler import _make_handler

cdef extern from "pyref.h":
    cdef cppclass PyRef:
        PyRef(PyObject* ref_)
        PyObject* ref


cdef logger = logging.getLogger("larch.reactive")

#some gevent support
cdef extern from "greenlet.h":
    ctypedef PyObject PyGreenlet

    PyGreenlet* PyGreenlet_GetCurrent()
    void PyGreenlet_Import()


ctypedef long (*get_fiber_id_t)()
cdef get_fiber_id_t get_current_fiber_id

ctypedef object (*get_fiber_t)()
cdef get_fiber_t get_current_fiber


cdef c_current_thread = None

cdef object get_current_thread():
    return c_current_thread()

cdef long get_current_greenlet_id():
    cdef PyGreenlet* g = PyGreenlet_GetCurrent()
    Py_DECREF(<object>g)
    return <long>g


try:
    from greenlet import GreenletExit, getcurrent
except ImportError:
    from threading import current_thread

    class GreenletExit(BaseException):
        pass

    c_current_thread = current_thread
    get_current_fiber = get_current_thread
    get_current_fiber_id = <get_fiber_id_t>PyThread_get_thread_ident
else:
    PyGreenlet_Import()

    get_current_fiber = <get_fiber_t>PyGreenlet_GetCurrent
    get_current_fiber_id = get_current_greenlet_id


__all__ = ("rcontext", "CellBase", "ReactiveState", "Subject", "Container",
           "Rule", "IterRule", "ResetContainer", "pointer_attrgetter",
           "pointer_itemgetter", "pointer_resolve", "ReactiveWarning")


cdef class Subject
cdef class Observer
cdef class Rule
cdef class Container
cdef class ReactiveContext
cdef class ReactiveState
cdef class CellBase


cdef class DecoratorContext:
    def call(self, func, *args, **kwargs):
        with self:
            return func(*args, **kwargs)

    def __call__(self, func):
        result = update_wrapper(_make_handler(self, func), func)
        result.decorated = func
        return result


cdef class AtomicDecoratorContext(DecoratorContext):
    def __enter__(self):
        ircontext._atomic_start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        ircontext._atomic_end()
        return False


cdef class SilentDecoratorContext(DecoratorContext):
    cdef int old_do_notify

    def __enter__(self):
        cdef _ContextVars v = ircontext._vars()
        ircontext._atomic_start()
        self.old_do_notify = v.do_notify
        v.do_notify = 0

    def __exit__(self, exc_type, exc_val, exc_tb):
        ircontext._vars().do_notify = self.old_do_notify
        ircontext._atomic_end()
        return False


cdef class UntouchedDecoratorContext(DecoratorContext):
    cdef int old_do_touch

    def __enter__(self):
        cdef _ContextVars v = ircontext._vars()
        self.old_do_touch = v.do_touch
        v.do_touch = 0

    def __exit__(self, exc_type, exc_val, exc_tb):
        ircontext._vars().do_touch = self.old_do_touch
        return False


cdef class TouchedDecoratorContext(DecoratorContext):
    cdef int old_do_touch

    def __enter__(self):
        cdef _ContextVars v = ircontext._vars()
        self.old_do_touch = v.do_touch
        v.do_touch = 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        ircontext._vars().do_touch = self.old_do_touch
        return False


ctypedef deque[PyRef] ObjectDeque
ctypedef deque[PyRef].iterator ObjectDequeIter


cdef class _ContextVars:
    cdef:
        public long atomic_start_round
        public long _call_counter
        public int atomic_count
        public object current_observer
        public int do_touch
        public int do_notify
        public dict old_values
        public list observers
        public dict observer_count
        ObjectDeque* callbacks

    def __init__(self):
        #for a documentation of the attributes
        #see pcore.py
        self.atomic_start_round = 0
        self.atomic_count = 0
        self._call_counter = 0
        self.current_observer = None
        self.do_touch = <int> 0
        self.do_notify = <int>1
        self.old_values = {}
        self.observers = []
        self.observer_count = {}
        self.callbacks = new ObjectDeque()

    def __dealloc__(self):
        del self.callbacks


class ReactiveWarning(UserWarning):
    pass


cdef class ReactiveContext:
    cdef:
        public AtomicDecoratorContext atomic
        long _last_fiber_id
        _ContextVars _current_vars

    def __init__(self):
        self.atomic = AtomicDecoratorContext()
        self._last_fiber_id = 0

    cpdef _ContextVars _vars(self):
        cdef long fiber_id = get_current_fiber_id()

        if fiber_id == self._last_fiber_id:
            return self._current_vars

        self._last_fiber_id = fiber_id
        fiber = get_current_fiber()

        rvars = getattr(fiber, "__larch_reactive_vars__", None)
        if rvars is None:
            rvars = fiber.__larch_reactive_vars__ = _ContextVars()

        self._current_vars = <_ContextVars>rvars
        return self._current_vars

    property observers:
        def __get__(self):
            return self._vars().observers

    property old_values:
        def __get__(self):
            return self._vars().old_values

    property untouched:
        def __get__(self):
            return UntouchedDecoratorContext()

    property touched:
        def __get__(self):
            return TouchedDecoratorContext()

    property silent:
        def __get__(self):
            return SilentDecoratorContext()

    property rounds:
        def __get__(self):
            return self._vars().atomic_start_round

    property transaction_level:
        def __get__(self):
            return self._vars().atomic_count

    property inside_rule:
        def __get__(self):
            return self._vars().current_observer

    property inside_touch:
        def __get__(self):
            cdef _ContextVars x = self._vars()
            return bool(x.do_touch)

    cpdef int _atomic_start(self) except -1:
        cdef _ContextVars x = self._vars()
        x.atomic_start_round = x._call_counter
        x.atomic_count += 1
        return 0

    cpdef int _atomic_end(self) except -1:
        cdef:
            list observers
            dict observer_count
            Observer observer
            _ContextVars x = self._vars()
            size_t count
            int greenlet_exit = 0

        if x.atomic_count != 1:
          x.atomic_count -= 1
          return 0

        observer_count = x.observer_count
        observers = x.observers

        try:
          for i in range(10000):
              if not observers:
                  break
              observer = observers.pop()

              count = observer_count.get(observer, 0)
              if count <= 1:
                  del observer_count[observer]
                  try:
                      observer.notify()
                  except GreenletExit:
                      greenlet_exit = 1
                      logger.exception(
                          "GreenletExit during observer notification",
                          stack_info=True)
              else:
                  observer_count[observer] = count - 1
          else:
              x.callbacks.clear()
              exp = RuntimeError("possible endless recursion {0}".format(observers))
              del observers[:]
              observer_count.clear()
              raise exp
        finally:
          x.old_values.clear()
          x.atomic_count = 0

        if not x.callbacks.empty():
            greenlet_exit = self._execute_callbacks(x, greenlet_exit)

        if greenlet_exit:
          raise GreenletExit()

        return 0

    cdef int _execute_callbacks(self, _ContextVars x,
                                int greenlet_exit) except -1:
        cdef:
            ObjectDeque* callbacks
            ObjectDequeIter it

        callbacks = x.callbacks
        x.callbacks = new ObjectDeque()

        it = callbacks.begin()
        while it != callbacks.end():
            c = <object>deref(it).ref
            inc(it)
            try:
                c()
            except Exception as e:
                from .exception import format_full_exception_tb
                tb = format_full_exception_tb()
                exc_type, exc_value, exc_traceback = sys.exc_info()
                warnings.warn(
                    'Exception while executing callback: "{0}"\n{1}'.format(
                        exc_value, tb), ReactiveWarning)
                logger.exception("Exception while executing callback %r: %r",
                                  c, e, stack_info=True)
            except GreenletExit:
                greenlet_exit = 1
                logger.exception("GreenletExit while executing callback %r", c, stack_info=True)

        del callbacks
        return greenlet_exit

    cpdef int rule_call(self, Rule rule, args, kwargs) except -1:
        cdef _ContextVars x = self._vars()
        old_do_touch = x.do_touch
        old_observer = x.current_observer
        x.current_observer = rule
        x.do_touch = 1
        try:
            rule.call_method(args, kwargs)
        except GreenletExit:
            raise
        except BaseException as e:
            from .exception import format_full_exception_tb
            tb = format_full_exception_tb()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            msg = ('Exception while calling rule "{0}({1!r}, {2!r})": "{3!r}"\n{4}')
            warnings.warn(msg.format(rule, args, kwargs, exc_value, tb), ReactiveWarning)
            logger.exception("Exception while calling rule %r", e, stack_info=True)
        finally:
            x.do_touch = old_do_touch
            x.current_observer = old_observer

    cpdef int notify(self, Subject subject, old_value) except *:
        cdef _ContextVars x = self._vars()
        if x.do_notify:
            subject_id = id(subject)
            x.old_values.setdefault(subject_id, old_value)
            self.emit(subject.get_observers())
            subject.clear_observers()

    cpdef int emit(self, list observers) except *:
        cdef:
            _ContextVars x = self._vars()
            size_t count

        for o in observers:
            insort_left(x.observers, o)
            count = x.observer_count.get(o, 0)
            x.observer_count[o] = count + 1

    cpdef int push_callback(self, callback) except -1:
        cdef _ContextVars x = self._vars()
        if x.atomic_count:
            x.callbacks.push_back(PyRef(<PyObject*>callback))
        else:
            callback()

    cdef int shift_callback(self, callback) except -1:
        cdef _ContextVars x = self._vars()
        if x.atomic_count:
            x.callbacks.push_front(PyRef(<PyObject*>callback))
        else:
            callback()

    cpdef long call_counter(self):
        cdef _ContextVars x = self._vars()
        x._call_counter += 1
        return x._call_counter

    cpdef object touch(self, Subject subject):
        cdef _ContextVars x = self._vars()
        if x.do_touch:
            subject.add_observer(x.current_observer)
            return True

        return False


cdef ReactiveContext ircontext = ReactiveContext()
rcontext = ircontext


def null_func():
    pass


cdef class Observer:
    cdef:
        public size_t last_call
        public int level

    cpdef notify(self):
        pass

    def __hash__(self):
        return <int><long><PyObject*>self

    cdef int is_less(self, Observer other):
        # reverse
        return (self.level > other.level
                or self.level==other.level and self.last_call > other.last_call)


cdef class Rule(Observer):
    """A rule that depends on cells an manipulate cells"""
    cdef public object holder
    cdef public method
    cdef object __weakref__

    def __init__(self, holder, method, level, last_call):
        self.holder = holder
        self.method = method
        self.last_call = last_call
        self.level = level

    cdef int call_method(self, args, kwargs) except -1:
        holder = <object>PyWeakref_GET_OBJECT(self.holder)
        if holder is not None:
            try:
                self.method(holder, *args, **kwargs)
            except StopIteration:
                pass

    cpdef notify(self, args=(), kwargs={}):
        self.last_call = ircontext.call_counter()
        ircontext.rule_call(self, args, kwargs)

    def __repr__(self):
        holder = <object>PyWeakref_GET_OBJECT(self.holder)
        if holder is None:
            return "<Rule obsolete {0} ({1}, {2})>".format(
                self.method.__name__, self.level, self.last_call)

        return "<Rule {0}({1}, {2}) of {3!r}>".format(
            self.method.__name__, self.level, self.last_call, self.holder())


cdef class _IterCaller:
    cdef object iterator

    def __init__(self, iterator):
        self.iterator = iterator

    def call(self):
        try:
            next(self.iterator)
            rcontext.push_callback(self.call)
        except StopIteration:
            pass
        except RuntimeError as e:
            if e.args != ('generator raised StopIteration',):
                raise


cdef class IterRule(Rule):
    cdef int call_method(self, args, kwargs) except -1:
        cdef _IterCaller caller
        holder = <object>PyWeakref_GET_OBJECT(self.holder)
        if holder is not None:
            iterator = self.method(holder, *args, **kwargs)
            caller = _IterCaller(iterator)
            caller.call()


cdef class Subject:
    """An observer's subject"""
    cdef public dict _observers
    cdef public size_t _last_counter

    def __init__(self):
        self._observers = {}
        self._last_counter = 0

    cpdef int touch(self):
        ircontext.touch(self)

    cpdef int changed(self, object old_value=None) except -1:
        if self._observers:
            ircontext._atomic_start()
            ircontext.notify(self, old_value)
            ircontext._atomic_end()

        return 0

    cpdef int take(self, Subject other):
        self._observers.update(other._observers)
        self._last_counter = max(self._last_counter, other._last_counter)

    cpdef list get_observers(self):
        return self.filter_observers(self._observers)

    cdef list filter_observers(self, dict observers):
        cdef:
            size_t last_counter = self._last_counter
            Observer o
            list result = []

        for r in observers.values():
            o = <Observer>PyWeakref_GET_OBJECT(r)
            if o is not None and o.last_call <= last_counter:
                result.append(o)

        return result

    cpdef int add_observer(self, Observer observer) except -1:
        if observer:
            self._last_counter = max(observer.last_call, self._last_counter)
            self._observers[id(observer)] = PyWeakref_NewRef(observer, None)
        return 0

    cpdef int clear_observers(self) except -1:
        self._observers = {}
        return 0


cdef class ReactiveState:
    cdef tuple containers

    def __init__(self, reactive):
        cdef:
            CellBase c
            dict cells = reactive.__reactive_cells__

        self.containers = <tuple>PyTuple_New(len(cells))
        for c in cells.values():
            tmp = c.create_container(reactive)
            Py_INCREF(tmp)
            PyTuple_SET_ITEM(self.containers, c.index, tmp)

    cdef void replace_container(self, int index, Container container):
        cdef Container tmp = <Container>self.get(index)
        Py_DECREF(tmp)
        Py_INCREF(container)
        PyTuple_SET_ITEM(self.containers, index, container)

    cdef inline Container get(self, int index):
        return <Container>PyTuple_GET_ITEM(self.containers, index)

    def from_dict(self, reactive, state):
        cdef:
            CellBase c
            Container tmp
            dict cells = reactive.__reactive_cells__

        for c in cells.values():
            if c.name in state:
                self.get(c.index)._value = state.get(c.name)

    def as_dict(self, reactive):
        cdef:
            CellBase c
            Container tmp
            dict cells = reactive.__reactive_cells__

        result = {}
        for c in cells.values():
            result[c.name] = self.get(c.index)._value

        return result


cdef class Container(Subject):
    """A cell container that actually contains the value"""
    cdef public object _value

    def __init__(self, value):
        super(Container, self).__init__()
        self._value = value
        #A destroyed container can be recreated (i.e. it has the same id)
        #in the same atomic round providing a wrong old_value for
        #the new (recreated) Container leading to wrong behaviour
        ircontext._vars().old_values.pop(id(self), None)

    cpdef int set_value(self, object value) except *:
        cdef object old_value = self._value
        self._value = value
        if old_value != value or type(old_value) != type(value):
            self.changed(old_value)
            return 1
        return 0

    cpdef object get_value(self):
        self.touch()
        return self._value

    def __repr__(self):
        return "<{0}({1:x}: {2!r}>".format(
            self.__class__.__name__, id(self), self._value)


cdef class _ResetValue:
    cdef public Container dest
    cdef public object old_value

    def __cinit__(self, dest, old_value):
        self.dest = dest
        self.old_value = old_value

    cpdef reset(self):
        self.dest._value = self.old_value


cdef class ResetContainer(Container):
    cdef public object _reset_value

    def __init__(self, value):
        super(ResetContainer, self).__init__(value)
        self._reset_value = value

    cpdef int set_value(self, value) except *:
        cdef:
            _ResetValue r
            object old_value = self._value

        self._value = value
        if old_value != value or type(old_value) != type(value):
            r = _ResetValue(self, self._reset_value)
            ircontext._atomic_start()
            self.changed(self._reset_value)
            ircontext.shift_callback(r.reset)
            ircontext._atomic_end()
            return 1

        return 0

cdef class CellBase:
    cdef public int index
    cdef public object name

    def __init__(self):
        self.name = ""

    def __init_cell__(self, name, index):
        self.name = name
        self.index = index

    def __get__(self, object holder, object owner):
        cdef ReactiveState state

        if holder is None:
            return self

        state = holder.__reactive_state__
        return state.get(self.index).get_value()

    def __set__(self, object holder, object value):
        cdef ReactiveState state = holder.__reactive_state__
        state.get(self.index).set_value(value)

    cpdef object get_container(self, object holder):
        cdef ReactiveState state = holder.__reactive_state__
        return state.get(self.index)

    cpdef int set_container(self, object holder, object value):
        #if we replace the container we have to keep the rules
        cdef ReactiveState state = holder.__reactive_state__
        value.take(state.get(self.index))
        state.replace_container(self.index, value)


cdef class pointer_accessor:
    cdef object get(self, obj):
        return None

    def __call__(self, obj):
        return self.get(obj)


cdef inline bytes to_bytes3(name):
    return bytes(name, "utf-8")

cdef inline bytes to_bytes2(name):
    return bytes(name)

ctypedef bytes (*to_bytes_t)(name)
cdef to_bytes_t to_bytes

try:
    to_bytes3("t")
except TypeError:
    to_bytes = to_bytes2
else:
    to_bytes = to_bytes3


cdef class pointer_attrgetter(pointer_accessor):
    cdef public bytes name

    def __init__(self, name):
        self.name = to_bytes(name)

    cdef object get(self, obj):
        cdef char* name = PyBytes_AS_STRING(self.name)
        return PyObject_GetAttrString(obj, name)


cdef class pointer_itemgetter(pointer_accessor):
    cdef public index

    def __init__(self, index):
        self.index = index

    cdef object get(self, obj):
        return obj[self.index]


def pointer_resolve(obj, tuple accessors):
    cdef pointer_accessor a

    for a in accessors:
        obj = a.get(obj)

    return obj


cdef insort_left(list olist, Observer item):
    cdef:
        size_t lo = 0
        size_t hi = len(olist)
        size_t mid
        Observer litem
        int res

    while lo < hi:
        mid = (lo + hi) >> 1
        litem = <Observer>PyList_GET_ITEM(olist, mid)
        if litem.is_less(item):
            lo = mid + 1
        else:
            hi = mid

    olist.insert(lo, item)



"""
Provides a pointer to an attribute
"""
import sys
import operator
from weakref import ref
# __pragma__ ('skip')
from .core import rcontext, pointer_attrgetter, pointer_itemgetter, pointer_resolve
__all__ = ("PointerBase", "Pointer", "PointerMap", "PointerState", "PointerExpression", "SELF",
           "ResolveError", "merge_pointers", "NOTHING")
# __pragma__ ('noskip')
# __pragma__ ('ecom')
"""?
from .pcore import rcontext, pointer_attrgetter, pointer_itemgetter, pointer_resolve
?"""
# __pragma__ ('noecom')


def _check_exception(exc, pointer_state, exception_type, to_transfrom):
    # __pragma__ ('skip')
    type_, _, traceback = sys.exc_info()
    if isinstance(exc, to_transfrom):
        raise exception_type(repr(pointer_state), exc).with_traceback(traceback)

    exc.args += (repr(pointer_state),)
    raise exc.with_traceback(traceback)
    # __pragma__ ('noskip')
    """?
    console.error("Exception in pointer", repr(exc))
    if isinstance(exc, to_transfrom):
        raise exception_type()
    raise exc
    ?"""


def _check_setter_exception(exc, pointer_state, exception_type):
    _check_exception(exc, pointer_state, exception_type,
                     (AttributeError, IndexError, KeyError, ReferenceError))


def _check_getter_exception(exc, pointer_state, exception_type):
    _check_exception(exc, pointer_state, exception_type,
                     (AttributeError, TypeError, IndexError, KeyError, ReferenceError))


class MetaNothing(type):
    def __init__(self, name, bases, dict_):
        super().__init__(name, bases, dict_)

    def __bool__(self):
        return False


class NOTHING(metaclass=MetaNothing):
    pass


_RNOTHING = ref(NOTHING)


# __pragma__ ('opov')
class ResolveError(ValueError):
    def __init__(self, path, original):
        self.path = path
        self.original = original

    def __repr__(self):
        return f"<{self.__class__.__name__} path={self.path} original={repr(self.original)}>"


class StrongRef:
    __slots__ = ("ref",)

    def __init__(self, ref):
        self.ref = ref

    def __call__(self):
        return self.ref


def create_ref(obj):
    try:
        return ref(obj)
    except TypeError:
        return StrongRef(obj)


class PointerState:
    __slots__ = ("root", "path", "accessors")

    def __init__(self, root=_RNOTHING, path=(), accessors=()):
        # root must be a weak reference or a StrongRef
        self.root = root
        self.path = path
        self.accessors = accessors

    def __eq__(self, other):
        if other is not None and self.__class__ == other.__class__:
            return self.root == other.root and self.path == other.path
        return False

    def __ne__(self, other):
        return not (self == other)

    def __bool__(self):
        try:
            self.get()
            return False if self.root is _RNOTHING else True
        except ResolveError:
            return False

    def __len__(self):
        return len(self.accessors)

    def __hash__(self):
        return hash(self.path)

    def new_item(self, index):
        path = self.path + (index,)
        accessors = self.accessors + (pointer_itemgetter(index), )
        return self.create_state(self.root, path, accessors)

    def new_attr(self, name):
        path = self.path + (name,)
        accessors = self.accessors + (pointer_attrgetter(name), )
        return self.create_state(self.root, path, accessors)

    def create_state(self, root, path, accessors):
        return self.__class__(root, path, accessors)

    def get(self):
        return self.delegate_get(self.root())

    def delegate_get(self, root):
        try:
            return pointer_resolve(root, self.accessors)
        except Exception as e:
            _check_getter_exception(e, self, ResolveError)

    def set(self, value, root=None):
        parent, getter, setter = self.split(root)
        try:
            setter(parent, value)
        except Exception as e:
            _check_setter_exception(e, self, ResolveError)
        return value

    def split(self, root=None):
        def setitem(obj, value):
            obj[self.path[-1]] = value

        def setattr_(obj, value):
            setattr(obj, self.path[-1], value)
            return value

        if root is None:
            root = self.root()

        # __pragma__ ('tconv')
        if not self.accessors:
            def invalid_setter(obj, value):
                raise TypeError("cannot set")

            return root, (lambda i: i), invalid_setter
        # __pragma__ ('notconv')

        setter = setitem if type(self.accessors[-1]) is pointer_itemgetter else setattr_
        try:
            parent = pointer_resolve(root, self.accessors[:-1])
        except Exception as e:
            _check_getter_exception(e, self, ResolveError)

        return parent, self.accessors[-1], setter

    def is_same(self, other):
        """returns True if other and self refers to the same object"""
        me = self
        try:
            if len(me.path) < len(other.path):
                other = other.sub_state(len(other.path) - len(me.path))
            else:
                me = me.sub_state(len(me.path) - len(other.path))
        except ResolveError:  # pragma: no cover
            return False

        def tmt(a):
            return tuple(map(type, a))

        if isinstance(me.root, ref):
            if not isinstance(other.root, ref) or me.root() != other.root():
                return False
        elif me.root != other.root:
            return False

        return (me.path == other.path
                and tmt(me.accessors) == tmt(other.accessors))

    def sub_state(self, count):
        """returns a state resolved to count path elements"""
        if not count:
            return self

        access = self.accessors
        try:
            obj = pointer_resolve(self.root(), access[:count])
            return self.create_state(
                ref(obj), self.path[count:], access[count:])
        except Exception as e:  # pragma: no cover
            _check_getter_exception(e, self, ResolveError)

    def merge_to(self, state):
        return self.create_state(state.root, state.path + self.path,
                                 state.accessors + self.accessors)

    def __getstate__(self):
        return self.root(), self.path, self.accessor_to_pickle()

    def __setstate__(self, state):
        root, self.path, accessors = state
        self.root = _RNOTHING if root is NOTHING else create_ref(root)
        self.pickle_to_accessor(accessors)

    def accessor_to_pickle(self):
        mapper = {pointer_itemgetter: "i", pointer_attrgetter: "a"}
        return "".join(mapper[type(a)] for a in self.accessors)

    def pickle_to_accessor(self, state):
        mapper = {"i": pointer_itemgetter, "a": pointer_attrgetter}
        self.accessors = tuple(mapper[g](a) for g, a in zip(state, self.path))

    def __str__(self):
        try:
            return "".join(
                f"[{name}]" if type(accessor) == pointer_itemgetter else f".{name}"
                for accessor, name in zip(self.accessors, self.path))
        except Exception as e:  # pragma: no cover
            e.args += (self.path, )
            raise e

    def __repr__(self):
        return f"{repr(self.root())}{str(self)}"


class _DumyState:
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.value == other.value

    def __repr__(self):
        return str(self.value)

    def get(self):
        return self.value

    def delegate_get(self, root):
        return self.value

    def merge_to(self, state):
        return self.value


def dumy_state(a):
    return a if isinstance(a, PointerState) else _DumyState(a)


class _ExpressionRoot:
    # root for PointerExpression
    __slots__ = ("args", )

    def __init__(self, args):
        self.args = tuple(a.__state__ if isinstance(a, Pointer) else dumy_state(a) for a in args)

    def __eq__(self, other):
        if other is not None and self.__class__ == other.__class__:
            return self.args == other.args
        return False

    def merge_args(self, state):
        return (a.merge_to(state) for a in self.args)

    def merge_to(self, state):
        return self.__class__(self.merge_args(state))

    def __getstate__(self):
        return self.args

    def __setstate__(self, state):
        self.args = state


class _ExpressionOperandRoot(_ExpressionRoot):
    __slots__ = ("operand", "kwargs")

    def __init__(self, operand, args, kwargs):
        super().__init__(args)
        self.operand = operand
        self.kwargs = {
            k: v.__state__ if isinstance(v, Pointer) else _DumyState(v)
            for k, v in kwargs.items()}

    def __eq__(self, other):
        if super().__eq__(other):
            return self.operand == other.operand
        return False

    def __call__(self):
        args = [a.get() for a in self.args]
        kwargs = {k: v.get() for k, v in self.kwargs.items()}
        return self.apply(args, kwargs)

    def evaluate(self, root):
        # like call but with different root in argument
        args = [a.delegate_get(root) for a in self.args]
        kwargs = {k: v.delegate_get(root) for k, v in self.kwargs.items()}
        return self.apply(args, kwargs)

    def apply(self, args, kwargs):
        return self.operand(*args, **kwargs)

    def __repr__(self):
        return f"{self.operand.__name__}{self.args}{self.kwargs}"

    def merge_kwargs(self, state):
        return {k: v.merge_to(state) for k, v in self.kwargs.items()}

    def merge_to(self, state):
        return self.__class__(self.operand, self.merge_args(state),
                              self.merge_kwargs(state))

    def __getstate__(self):
        return self.operand, self.args, self.kwargs

    def __setstate__(self, state):
        try:
            self.operand, self.args, self.kwargs = state
        except ValueError:
            self.operand, self.args = state
            self.kwargs = {}


class _ExpressionOperandRootApply(_ExpressionOperandRoot):
    def apply(self, args, kwargs):
        return self.operand(args)


class _ExpressionAndRoot(_ExpressionRoot):
    def __call__(self):
        return self.apply(lambda v: v.get())

    def evaluate(self, root):
        return self.apply(lambda v: v.delegate_get(root))

    def apply(self, getter):
        try:
            values = (getter(a) for a in self.args)
            v = False
            # __pragma__ ('tconv')
            for v in values:
                if not v:
                    return False
            # __pragma__ ('notconv')
            return v
        except Exception:
            return False

    def __repr__(self):
        return f"and{self.args}"


class _ExpressionOrRoot(_ExpressionRoot):
    def __call__(self):
        return self.apply(lambda v: v.get())

    def evaluate(self, root):
        return self.apply(lambda v: v.delegate_get(root))

    def apply(self, getter):
        def eval_(v):
            try:
                v = getter(v)
            except Exception:
                return False
            return v

        values = (eval_(a) for a in self.args)
        # __pragma__ ('tconv')
        for v in values:
            if v:
                return v
        # __pragma__ ('notconv')
        return v

    def __repr__(self):
        return f"or{self.args}"


class PointerExpression(PointerState):
    """A pointer expression"""

    def set(self, value, root=None):
        raise ValueError("Cannot set PointerExpression", self)

    def get(self):
        return super().delegate_get(self.root())

    def delegate_get(self, root):
        return super().delegate_get(self.root.evaluate(root))

    def merge_to(self, state):
        root = self.root.merge_to(state)
        return self.create_state(root, self.path, self.accessors)

    # __pragma__ ('kwargs')
    @classmethod
    def make(cls, pointer, operand, *args, **kwargs):
        return pointer.__class__(cls(_ExpressionOperandRoot(operand, args, kwargs)))
    # __pragma__ ('nokwargs')

    @classmethod
    def make_and(cls, pointer, *args):
        return pointer.__class__(cls(_ExpressionAndRoot(args)))

    @classmethod
    def make_or(cls, pointer, *args):
        return pointer.__class__(cls(_ExpressionOrRoot(args)))

    # __pragma__ ('kwargs')
    @classmethod
    def call(cls, func, *args, **kwargs):
        """calls func with args"""
        return Pointer(cls(_ExpressionOperandRoot(func, args, kwargs)))
    # __pragma__ ('nokwargs')

    @classmethod
    def apply(cls, func, *args):
        return Pointer(cls(_ExpressionOperandRootApply(func, args, {})))

    def __getstate__(self):
        return self.root, self.path, self.accessor_to_pickle()

    def __setstate__(self, state):
        self.root, self.path, accessors = state
        self.pickle_to_accessor(accessors)

    def __repr__(self):
        return f"{repr(self.root)}{self}"


class PointerBase:
    __slots__ = ()


class Pointer(PointerBase):
    """
    A placeholder to an object attributes and index chain (a generalization to
    attrgetter/itemgetter of module operator) to construct a pointer you call

    >>> pointer = Pointer(obj).a[0]

    to get a pointer's value you call

    >>> value = pointer()  # same as obj.a[0]

    to set a pointer's value you call

    >>> pointer(value)  # same as obj.a[0] = value
    """
    __slots__ = ("__state__", "__weakref__")

    def __init__(self, state_or_obj=None):
        if state_or_obj is None:
            state_or_obj = PointerState()

        elif isinstance(state_or_obj, Pointer):
            state_or_obj = state_or_obj.__state__

        elif not isinstance(state_or_obj, PointerState):
            state_or_obj = PointerState(create_ref(state_or_obj))

        self.__state__ = state_or_obj

    def __get__(self, holder, owner):
        # delegator handler
        if holder is None:
            return self

        return self.__state__.delegate_get(holder)

    def __set__(self, holder, value):
        # delegator handler
        self.__state__.set(value, holder)

    def __bool__(self):
        return bool(self.__state__)

    def __getitem__(self, index):
        return self.__class__(self.__state__.new_item(index))

    def __getattr__(self, name):
        return self.__class__(self.__state__.new_attr(name))

    def __call__(self, value=NOTHING):
        return self.__state__.get() if value is NOTHING else self.__state__.set(value)

    def __hash__(self):
        return hash(self.__state__)

    def __eq__(self, other):
        if other is not None and self.__class__ == other.__class__:
            return self.__state__.is_same(other.__state__)

        return False

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        with rcontext.untouched:
            return str(self.__state__)

    def __len__(self):
        return 0

    def __iter__(self):
        raise TypeError("cannot iter %r" % self)

    def __repr__(self):
        return f"<{self.__class__.__name__}-{repr(self.__state__)}>"

    def __add__(self, other):
        return PointerExpression.make(self, operator.add, self, other)

    def __radd__(self, other):
        return PointerExpression.make(self, operator.add, other, self)

    def __sub__(self, other):
        return PointerExpression.make(self, operator.sub, self, other)

    def __rsub__(self, other):
        return PointerExpression.make(self, operator.sub, other, self)

    def __truediv__(self, other):
        return PointerExpression.make(self, operator.truediv, self, other)

    def __rtruediv__(self, other):
        return PointerExpression.make(self, operator.truediv, other, self)

    def __mul__(self, other):
        return PointerExpression.make(self, operator.mul, self, other)

    def __rmul__(self, other):
        return PointerExpression.make(self, operator.mul, other, self)

    def __and__(self, other):
        return PointerExpression.make_and(self, self, other)

    def __rand__(self, other):
        return PointerExpression.make_and(self, other, self)

    def __or__(self, other):
        return PointerExpression.make_or(self, self, other)

    def __ror__(self, other):
        return PointerExpression.make_or(self, other, self)

    def __reduce_ex__(self, protocol):
        return self.__class__, (self.__state__,)


SELF = Pointer()
"""This is actually not a Pointer but a trivial delegator"""


class PointerMap:
    """A named value container which is able to handle pointer values"""
    __slots__ = ("__values", )

    # __pragma__ ('kwargs')
    def __init__(self, **kwargs):
        self.__values = kwargs
    # __pragma__ ('nokwargs')

    def get(self, key, default=NOTHING):
        value = self.__values.get(key, default)
        if isinstance(value, PointerBase):
            return value()

        elif value is NOTHING:
            raise KeyError(key)

        return value

    def set(self, key, value):
        """Returns True if the key was a pointer"""
        val = self.__values.get(key, None)
        if isinstance(val, PointerBase) and not isinstance(value, PointerBase):
            val(value)
            return True

        self.__values[key] = value
        return False

    def remove(self, key):
        self.__values.pop(key, None)

    def update(self, dictionary):
        self.__values.update(dictionary)

    def keys(self):
        return self.__values.keys()

    def raw(self, key):
        return self.__values.get(key)

    def __contains__(self, key):
        return key in self.__values


def merge_pointers(*pointers):
    """
    Merges several pointers together.

    Args:
        pointers: multiple pointers to merge

    Returns:
        Pointer: a chained pointer object.

    Example:
        >>> p1 = Pointer(obj1).a.b
        >>> p2 = Pointer(obj2).c.d
        >>> p3 = merge_pointers(p1, p2)
        >>> p3 == Pointer(obj1).a.b.c.d
    """
    states = (p.__state__ for p in pointers)
    state = next(states)
    for s in states:
        state = s.merge_to(state)

    return Pointer(state)



from __future__ import print_function
import sys
import os.path
import linecache
from contextlib import contextmanager
from .core import rcontext, CellBase
from .base import cells_of, SimpleReactive
from pprint import pprint
if sys.version_info < (3, 0):
    from cStringIO import StringIO
else:
    from io import StringIO


@rcontext.untouched
def dump(obj, connected_only=True, line_ending="\n"):
    """
    dumps all cells and their dependent observers to a text string.
    """
    lines = []
    cells = cells_of(obj)

    lines.append("cells of %r" % obj)
    for cell in cells:
        container = cell.get_container(obj)
        observers = container.get_observers()
        if connected_only and not observers:
            continue

        lines.append("  %s(%s-%x)" % (cell.name, cell.__class__.__name__, id(container)))
        lines.extend(sorted("    %r(%x)" % (o, id(o)) for o in observers))

    return line_ending.join(lines)


class RContextPatcher(object):
    indent_count = 0

    def __init__(self, output=sys.stderr):
        self.output = output
        if type(rcontext).__module__ != "larch.reactive.pcore":
            raise RuntimeError("Only python module support debug context")

    def start(self, touch=True, emit=True, rule=True):
        self._org_rule_call = rcontext.rule_call
        self._org_touch = rcontext._touch
        self._org_emit = rcontext.emit

        if rule:
            rcontext.rule_call = self.rule_call
        if touch:
            rcontext._touch = self._touch
            if rcontext.touch == self._org_touch:
                rcontext.touch = self._touch

        if emit:
            rcontext.emit = self.emit

    def stop(self):
        if rcontext.rule_call == self.rule_call:
            del rcontext.rule_call
        if rcontext._touch == self._touch:
            del rcontext._touch
        if rcontext.emit == self.emit:
            del rcontext.emit

        if rcontext.touch == self._touch:
            rcontext.touch = rcontext._touch

    def indent(self):
        return "  " * (rcontext.atomic_count - 1 + self.indent_count)

    def write_indented(self, text, indent=None):
        indent = indent or self.indent()
        for i in text.splitlines():
            self.output.write(indent+i+"\n")

    def write(self, *args):
        s = StringIO()
        print(*args, file=s)
        self.write_indented(s.getvalue())

    def pprint(self, obj, indent=0):
        s = StringIO()
        pprint(obj, stream=s)
        indent = self.indent() + " " * indent
        self.write_indented(s.getvalue(), indent)

    def rule_call(self, rule, args, kwargs):
        self.write("call rule", repr(rule))
        self.indent_count += 1
        self._org_rule_call(rule, args, kwargs)
        self.indent_count -= 1
        self.write("")

    @rcontext.untouched
    def _touch(self, subject):
        try:
            cell, holder = self.find_cell(subject)
        except TypeError:
            holder = "unknown"
            cell = subject

        self._org_touch(subject)
        self.write("touch ({}.{})".format(holder, cell))
        for relevant, f in self.frames():
            if not relevant:
                continue

            line = self.get_source_line(f)
            line = "  {} in {}({}): {!r}".format(
                f.f_code.co_name, f.f_code.co_filename, f.f_lineno, line)
            self.write(line)

            try:
                # traceback until rule is reached
                cls = type(f.f_locals["self"])
                method = getattr(cls, f.f_code.co_name)
                if method.__rule__:
                    break
            except (KeyError, AttributeError):
                pass

        self.write("")
        return True

    def emit(self, observers):
        line = ""
        for relevant, f in self.frames():
            if relevant:
                line = self.get_source_line(f)
                line = "{}({}): {!r}".format(f.f_code.co_filename, f.f_lineno, line)
                break

        self.write("emit({}) {}".format(rcontext.atomic_count, line))
        with rcontext.untouched:
            self.pprint(observers, 2)
        self._org_emit(observers)
        self.write("")

    def find_cell(self, subject):
        for r, f in self.frames():
            holder = f.f_locals.get("holder")
            cell = f.f_locals.get("self")
            if not r and isinstance(holder, SimpleReactive) and isinstance(cell, CellBase):
                container = cell.get_container(holder)
                if container is subject:
                    return cell, holder

    def frames(self):
        # yields the frames outside reactive package
        f = sys._getframe()
        rpath = os.path.split(__file__)[0]
        while f:
            path = os.path.split(f.f_code.co_filename)[0]
            yield path != rpath, f
            f = f.f_back

    def get_source_line(self, frame):
        lineno = frame.f_lineno
        co = frame.f_code
        filename = co.co_filename
        linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, frame.f_globals)
        if line:
            return line.strip()
        return ""


@contextmanager
def debug(touch=True, emit=True, rule=True, output=sys.stderr):
    patcher = RContextPatcher(output)
    patcher.start(touch, emit, rule)
    try:
        yield
    finally:
        patcher.stop()



from __future__ import print_function
import os
os.environ["LARCH_REACTIVE"] = "python"
import larch.reactive as ra
import larch.reactive.debug as dbg


class Test(ra.Reactive):
    a = ra.Cell()
    b = ra.Cell()
    error = ra.Cell()

    def print_a(self):
        print("a=", repr(self.a))

    @ra.rule
    def _rule_a(self):
        if self.a:
            self.print_a()

    @ra.rule
    def _rule_b(self):
        if self.b:
            with dbg.debug():
                print(self.b)
                self.a = 2

    @ra.rule
    def _rule_error(self):
        if self.error:
            raise RuntimeError()

t = Test()

print("debug rule a")
with dbg.debug():
    t.a = 1

print("------------------")
print("call rule b")
t.b = 1

print("------------------")
print("error")
t.error = 1



import sys
import warnings
from operator import itemgetter, attrgetter
from weakref import ref
# __pragma__ ('skip')
from functools import wraps
from inspect import isgeneratorfunction
from .core import rcontext, IterRule, Rule, CellBase, ReactiveState
__all__ = ("rule", "SimpleReactive", "MetaReactive", "Reactive", "cells_of",
           "atomic", "untouched", "touched", "silent", "reactive", "call_outside")
# __pragma__ ('noskip')
# __pragma__ ('ecom')
"""?
from .pcore import rcontext, IterRule, CellBase, ReactiveState
?"""


def atomic():
    """Returns an atomic decorator context"""
    return rcontext.atomic


def untouched():
    """Returns an untouched decorator context"""
    return rcontext.untouched


def touched():
    """Returns a touched decorator context"""
    return rcontext.touched


def silent():
    """Returns a slient decorator context"""
    return rcontext.silent


_rule_call_ = 0


def rule(level=0):
    """
    A decorator converting a method to a rule.

    Args:
        level (int): A "hard" priority level. Rules with smaller levels will
            allways be called before rules of greater levels.
    """

    global _rule_call_

    internal_rule_order = _rule_call_
    _rule_call_ += 1

    def irule(method, lvl=level, internal_order=internal_rule_order):
        # __pragma__ ('skip')
        @wraps(method)
        def call_rule(self, *args, **kwargs):
            if rcontext.inside_rule:
                return method(self, *args, **kwargs)

            # this is not done very often so we take the long
            # way to find the rule
            methods = (r for r in self.__reactive_rules__ if r.method is method)
            with atomic():
                try:
                    next(methods).notify(args, kwargs)
                except StopIteration:  # pragma: no cover
                    return method(self, *args, **kwargs)

        rule = IterRule if isgeneratorfunction(method) else Rule
        call_rule.__rule__ = (rule, lvl, internal_order, method)
        return call_rule
        # __pragma__ ('noskip')
        """?
        if not method.__rule__:
            method.__rule__ = (IterRule, lvl, internal_order, method)
        return method
        ?"""

    if not isinstance(level, int):
        return irule(level, 0)

    return irule


class SimpleReactive:
    """
    The base class for reactive objects that can hold cells and/or rules.
    Can also be used in multiple inheritance trees.
    """

    __reactive_state__ = None
    """An instance attribute holding the reference to a ReactiveState object"""

    __reactive_rules__ = None
    """
    An instance attribute, holding a tuple with all rule objects
    belonging to a reactive instance. This container is used, to prevent
    the rules objects from beeing deleted.
    """

    __reactive_rule_methods__ = None
    """A class attribute holding a tuple of all methods defined as rules"""

    __reactive_cells__ = None
    """A class attribute holding a dictionary of all cells with name"""

    @classmethod
    def __make_reactive__(cls):
        """converts an ordinary python class to a reactive python class"""
        cls.__replace_init__()
        cls.__init_reactive_attributes__()

    @classmethod
    def __init_reactive_attributes__(cls):
        cells, rules = cls.__collect_reactive_attributes__()
        cls.__reactive_rule_methods__ = tuple(rules)

        # Each class must have its own Cell Instances
        # to ensure an unqiue cell index
        d = dict(cls.__dict__)  # dict conversion for pyjs
        for i, (n, c) in enumerate(cells.items()):
            if n not in d:
                c = cells[n] = c.copy()

            c.__init_cell__(n, i)
            setattr(cls, n, c)

        cls.__reactive_cells__ = cells

    @classmethod
    def __collect_reactive_attributes__(cls):
        """collect all cells and rules of a class"""
        rules = []
        cells = {}

        for n in dir(cls):
            obj = getattr(cls, n)
            if isinstance(obj, CellBase):
                cells[n] = obj
            else:
                rule_props = getattr(obj, "__rule__", None)
                if isinstance(rule_props, tuple):
                    # factory, level, internal_order, method = rule_props
                    rules.append(rule_props)

        rules.sort(key=itemgetter(1, 2))
        return cells, rules

    @classmethod
    def __replace_init__(cls):
        """
        wraps the original __init__ function with a reactive __init__
        function that, initializes all rules and cells.
        """

        # __pragma__ ('kwargs')
        def _rinit_(self, *args, **kwargs):
            if self.__reactive_state__ is None:
                self.__reactive_state__ = ReactiveState(self)
                cls.__org_init__(self, *args, **kwargs)
                self.__init_rules__(self.__reactive_rule_methods__)
            else:
                cls.__org_init__(self, *args, **kwargs)
        # __pragma__ ('nokwargs')

        cls.__org_init__ = cls.__init__
        cls.__init__ = _rinit_  # __: skip
        # __pragma__('js', '{}', 'Object.defineProperty(cls, "__init__", {value: _rinit_});')

    def __init_rules__(self, rules):
        """Initializes all rules of a reactive object"""
        wself = ref(self)
        crules = [factory(wself, method, level, internal_order)
                  for factory, level, internal_order, method in rules]
        # __pragma__ ('tconv')
        if crules:
            self.__reactive_rules__ = crules
            with atomic():
                rcontext.emit(crules)
        # __pragma__ ('notconv')

    # __pragma__ ('kwargs')
    def __init__(self, *args, **kwargs):
        if kwargs:
            cell_names = self.__reactive_cells__.keys()
            cell_names = frozenset(cell_names).intersection(kwargs.keys())
            for n in cell_names:
                setattr(self, n, kwargs.pop(n))
        super().__init__(*args, **kwargs)
    # __pragma__ ('nokwargs')

    # __pragma__ ('skip')
    def __getstate__(self):
        state = {}
        state.update(self.__dict__)
        state.pop("__reactive_rules__", None)
        state["__reactive_state__"] = state["__reactive_state__"].as_dict(self)
        return state

    def __setstate__(self, state):
        cstate = {}  # copy state the original should not be changed
        cstate.update(state)

        rstate = cstate.pop("__reactive_state__")
        self.__dict__.update(cstate)

        self.__reactive_state__ = ReactiveState(self)
        self.__reactive_state__.from_dict(self, rstate)
        self.__init_rules__(self.__reactive_rule_methods__)

        # update the state a second time
        # because __init_rules__ could have change it accidently
        self.__reactive_state__.from_dict(self, rstate)
        self.__dict__.update(cstate)
    # __pragma__ ('noskip')


class MetaReactive(type):
    def __init__(self, name, bases, dict_):
        super().__init__(name, bases, dict_)
        self.__make_reactive__()


class Reactive(SimpleReactive, metaclass=MetaReactive):
    pass


def cells_of(reactive):
    """
    Returns all cell containers of an reactive object.

    Args:
        reactive: A reactive object.

    Returns:
        list: All reactive cell containers of `reactive`.
    """
    return sorted(reactive.__reactive_cells__.values(), key=attrgetter("order"))


# __pragma__ ('skip')
def reactive(cls):
    """A class decorator making a class reactive."""
    if not issubclass(cls, SimpleReactive):
        bases = (SimpleReactive, ) + cls.__bases__
        attribs = dict(cls.__dict__)
        cls = type(cls.__name__, bases, attribs)
    cls.__make_reactive__()
    return cls
# __pragma__ ('noskip')


def call_outside(func, *args, **kwargs):
    """
    Postpones the execution of `func` until the current atomic operation has
    finished. If called outside an atomic operation `func` is executed
    immediate.

    Args:
        func (callable): The callable to execute.
        *args: Arguments to pass to func
        **kwargs: Keyword arguments to pass to func

    Returns:
        callable: the given `func`
    """
    def caller(f=func, args=(args, kwargs)):
        try:
            f(*args[0], **args[1])
        except Exception:
            exc_type, exc_value, tb = sys.exc_info()
            # __pragma__ ('skip')
            from .exception import format_full_exception_tb
            tb = format_full_exception_tb()
            # __pragma__ ('noskip')
            tpl = 'Exception while executing outside call:  {}({}) "{}"\n{}'
            warnings.warn(tpl.format(f, args, exc_value, tb))
    caller.__name__ = func.__name__
    rcontext.push_callback(caller)
    return func
# __pragma__ ('noecom')



"""
Provides reactive collections
"""
from .core import ResetContainer


class MutableObserver(object):
    __slots__ = ("__action_start__", "__action_end__")

    def __init__(self):
        self.__action_start__ = ResetContainer(())
        self.__action_end__ = ResetContainer(())

    def __call__(self, value, method, *args, **kwargs):
        self.__action_start__.set_value(value)
        method(*args, **kwargs)
        self.__action_end__.set_value(value)


def _normalize_slice(list_, slice_):
    def normalize(index, default):
        if index is None:  # pragma: no cover
            index = default

        if index < 0:
            index += len(list_)

        return max(0, min(len(list_), index))

    return slice(normalize(slice_.start, 0), normalize(slice_.stop, len(list_)))


class ListBase(list):
    """
    The base list class which informs dependent rules if the list changes.
    """
    __slots__ = ("__observer__",)

    def __delitem__(self, key):
        if not isinstance(key, slice):
            key = slice(key, key + 1)

        key = _normalize_slice(self, key)
        if key.start >= len(self):
            # nothing to do
            return

        k = ("delete", key)
        self.__observer__(k, super(ListBase, self).__delitem__, key)

    def __delslice__(self, i, j):  # pragma: no cover
        self.__delitem__(slice(i, j))

    def remove(self, value):
        del self[self.index(value)]

    def __setitem__(self, key, value):
        if not isinstance(key, slice):
            if key < 0:
                key = len(self) + key

            key = slice(key, key + 1)
            value = [value]
        else:
            key = _normalize_slice(self, key)

        if not value:
            k = ("delete", key)
            if key.start == key.stop or key.start >= len(self):
                return  # no change at all
        else:
            k = ("change", (key, value))

        self.__observer__(k, super(ListBase, self).__setitem__, key, value)

    def __setslice__(self, i, j, value):  # pragma: no cover
        self.__setitem__(slice(i, j), value)

    def append(self, value):
        self.__observer__(("insert", (len(self), value)), super(ListBase, self).append, value)

    def insert(self, index, value):
        self.__observer__(("insert", (index, value)), super(ListBase, self).insert, index, value)

    def extend(self, value):
        self.__observer__(("extend", value), super(ListBase, self).extend, value)

    def pop(self, index=-1):
        if index < 0:
            index += len(self)

        value = self[index]
        self.__observer__(("delete", slice(index, index + 1)), super(ListBase, self).pop, index)
        return value

    def __iadd__(self, value):
        self.__observer__(("extend", value), super(ListBase, self).__iadd__, value)
        return self

    def __imul__(self, value):
        self.__observer__(("imul", value), super(ListBase, self).__imul__, value)
        return self

    def reverse(self):
        self.__observer__(("order", "reverse"), super(ListBase, self).reverse)

    def sort(self, *args, **kwargs):
        self.__observer__(("order", "sort"), super(ListBase, self).sort, *args, **kwargs)


class List(ListBase):
    """
    A list class which informs dependent rules before the list
    changes and after the change.
    """
    __slots__ = ("__observer__",)

    def __new__(cls, *args, **kwargs):
        obj = super(List, cls).__new__(cls, *args, **kwargs)
        obj.__observer__ = MutableObserver()
        return obj

    @property
    def __action_start__(self):
        return self.__observer__.__action_start__.get_value()

    @property
    def __action_end__(self):
        return self.__observer__.__action_end__.get_value()

    # __getsate__/__setstate__ to avoid __observer__ is pickled
    def __getstate__(self):
        return None

    def __setstate__(self, state):
        pass


class Dict(dict):
    __slots__ = ("__observer__",)

    def __new__(cls, *args, **kwargs):
        obj = super(Dict, cls).__new__(cls, *args, **kwargs)
        obj.__observer__ = MutableObserver()
        return obj

    @property
    def __action_start__(self):
        return self.__observer__.__action_start__.get_value()

    @property
    def __action_end__(self):
        return self.__observer__.__action_end__.get_value()

    # __getsate__/__setstate__ to avoid __observer__ is pickled
    def __getstate__(self):
        return None

    def __setstate__(self, state):
        pass

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)

        self.__observer__(("delete", key), super(Dict, self).__delitem__, key)

    def __setitem__(self, key, value):
        self.__observer__(("change", (key, value)),
                          super(Dict, self).__setitem__, key, value)

    def setdefault(self, key, default=None):
        if key not in self:
            self.__observer__(("change", (key, default)),
                              super(Dict, self).__setitem__, key, default)
        return self[key]

    def update(self, *args, **kwargs):
        self.__observer__(("update", (args, kwargs)),
                          super(Dict, self).update, *args, **kwargs)

    def popitem(self):
        for r in self.items():
            del self[r[0]]
            return r

        raise KeyError("empty")

    def pop(self, key, *args):
        if key in self:
            r = self[key]
            del self[key]
            return r

        if args:
            return args[0]

        raise KeyError(key)

    def clear(self):
        self.__observer__(("clear", None), super(Dict, self).clear)



from .version import __version__
from .base import (
    rule, SimpleReactive, MetaReactive, Reactive, cells_of, atomic, untouched, touched, silent,
    reactive, call_outside)
from .pointer import (
    PointerBase, Pointer, PointerMap, PointerState, PointerExpression, SELF, ResolveError,
    merge_pointers, NOTHING)
from .cell import Cell, MakeCell, ResetCell, TypeCell, MakeTypeCell
from .agent import Agent, CellAgent, OldAgent, old, cell
# __pragma__ ('skip')
from .core import (
    rcontext, CellBase, ReactiveState, Subject, Container, ResetContainer, Rule, IterRule,
    pointer_attrgetter, pointer_itemgetter, pointer_resolve, ReactiveWarning, core)

__all__ = (
    "__version__",
    "rcontext", "CellBase", "ReactiveState", "Subject", "Container", "ResetContainer", "Rule",
    "IterRule", "pointer_attrgetter", "pointer_itemgetter", "pointer_resolve", "ReactiveWarning",
    "core",

    "rule", "SimpleReactive", "MetaReactive", "Reactive", "cells_of", "atomic", "untouched",
    "touched", "silent", "reactive", "call_outside",

    "PointerBase", "Pointer", "PointerMap", "PointerState", "PointerExpression", "SELF",
    "ResolveError", "merge_pointers", "NOTHING",

    "Cell", "MakeCell", "ResetCell", "TypeCell", "MakeTypeCell",

    "Agent", "CellAgent", "OldAgent", "old", "cell")

# backward compatibility
ProxyBase = PointerBase
Proxy = Pointer
ProxyMap = PointerMap
ProxyState = PointerState
ProxyExpression = PointerExpression


def add_pickle_extensions():
    import copyreg
    from hashlib import md5

    def add_extension_by_hash(module, name):
        # prepare for the future
        code = int(md5((module + ":" + name).encode("ascii")).hexdigest(), 16)
        copyreg._inverted_registry[code] = (module, name)

        # backward compatible to old proxy module
        module_name = module.replace(".pointer", ".proxy") + ":" + name.replace("Pointer", "Proxy")
        code = int(md5((module_name).encode("ascii")).hexdigest(), 16)
        copyreg.add_extension(module, name, code & 0x7FFFFFFF)

    add_extension_by_hash("larch.reactive.pointer", "_ExpressionOperandRootApply")
    add_extension_by_hash("larch.reactive.pointer", "_ExpressionOperandRoot")
    add_extension_by_hash("larch.reactive.pointer", "_ExpressionAndRoot")
    add_extension_by_hash("larch.reactive.pointer", "_ExpressionOrRoot")
    add_extension_by_hash("larch.reactive.pointer", "_DumyState")
    add_extension_by_hash("larch.reactive.pointer", "PointerExpression")
    add_extension_by_hash("larch.reactive.pointer", "PointerState")
    add_extension_by_hash("larch.reactive.pointer", "Pointer")
    add_extension_by_hash("larch.reactive.pointer", "NOTHING")
# __pragma__ ('noskip')


# __pragma__ ('ecom')
"""?
from .pcore import (
    rcontext, CellBase, ReactiveState, Subject, Container, ResetContainer, Rule, IterRule,
    pointer_attrgetter, pointer_itemgetter, pointer_resolve, ReactiveWarning, core)
?"""
# __pragma__ ('noecom')



import sys
import traceback


def format_full_exception_tb():
    exc_info = sys.exc_info()
    tb = traceback.format_stack()[:-2]
    tb.extend(traceback.format_tb(exc_info[2]))
    tb.extend(traceback.format_exception_only(exc_info[0], exc_info[1]))
    tb_str = "Traceback (most recent call last):\n"
    tb_str += "".join(tb)[:-1]
    return tb_str



# depreceated module
from .pointer import (
    PointerBase, Pointer, PointerMap, PointerState, PointerExpression, SELF,
    ResolveError, merge_pointers, NOTHING, _DumyState, _ExpressionRoot, _ExpressionOperandRoot,
    _ExpressionOperandRootApply, _ExpressionAndRoot, _ExpressionOrRoot)

__all__ = ("NOTHING", "SELF", "ResolveError", "_DumyState", "_ExpressionRoot", "_ExpressionOrRoot",
           "_ExpressionOperandRoot", "_ExpressionOperandRootApply", "_ExpressionAndRoot")

ProxyBase = PointerBase
Proxy = Pointer
ProxyMap = PointerMap
ProxyState = PointerState
ProxyExpression = PointerExpression
merge_proxies = merge_pointers



"""
To use with coverage:
coverage erase
coverage run test_pcore.py
coverage combine
coverage report -m
"""
import os
os.environ["LARCH_REACTIVE"] = "python"
import unittest
from test_reactive import (RContextTest, PointerTest, AgentTest, CellTest,
                           TestReactive, MiscTest, ListTest, DictTest)
from test_thread import ThreadTest
from larch.reactive.core import core
try:
    import gevent
    from test_gevent import GeventTest
except ImportError:
    pass


class BackendTest(unittest.TestCase):
    def test_core(self):
        self.assertEqual(core, "python")


if __name__ == "__main__":
    unittest.main(failfast=True)



"""reactive agent"""
from .pointer import Pointer, ResolveError, _check_getter_exception
# __pragma__ ('skip')
from .core import rcontext, Container, CellBase
__all__ = ("Agent", "CellAgent", "OldAgent", "old", "cell")
# __pragma__ ('noskip')
# __pragma__ ('ecom')
"""?
from .pcore import rcontext, Container, CellBase
?"""
# __pragma__ ('noecom')


class Agent:
    def __init__(self, src):
        self.__src__ = src

    def __dir__(self):
        return self.__src__.__reactive_cells__.keys()

    def __getattr__(self, name):
        src = self.__src__
        cell = getattr(src.__class__, name)

        if isinstance(cell, Pointer):
            src, getter, setter = cell.__state__.split(src)
            try:
                cell = getter(src.__class__)
            except Exception:
                _check_getter_exception(cell.__state__, ResolveError)

            if not isinstance(cell, CellBase):
                raise TypeError("Attribute does not point to a cell", name)

        return cell.get_container(src)


class CellAgent(Agent):
    """returns cell containers as attributes"""

    def __setattr__(self, name, value):
        if name == "__src__":
            super().__setattr__(name, value)
        else:
            if isinstance(value, Container):
                cell = getattr(self.__src__.__class__, name)
                cell.set_container(self.__src__, value)
            else:
                raise TypeError("%r must be of type container" % value)


class OldAgent(Agent):
    """returns old cell values as attributes"""

    def __getattr__(self, name):
        try:
            container = super().__getattr__(name)
        except AttributeError:
            return getattr(self.__src__, name)

        # the default container.get_value() also touches the container
        return rcontext.old_values.get(id(container), container.get_value())


def _get_holder_cell(obj):
    holder, getter, setter = obj.__state__.split()
    try:
        return holder, getter(holder.__class__)
    except Exception as e:
        _check_getter_exception(e, obj.__state__, ResolveError)


def old(obj):
    """
    Returns an `old`-agent of `obj`. This function is helpfull within a rule,
    to access the old cell values before the current atomic operation.

    Args:
        obj (Reactive): A reactive object.

    Returns:
        An agent object offering the same attributes as `obj`, with the
        old cell values.
    """
    if isinstance(obj, Pointer):
        holder, cell = _get_holder_cell(obj)
        try:
            container = cell.get_container(holder)
        except AttributeError:
            return obj()
        return rcontext.old_values.get(id(container), container.get_value())
    else:
        return OldAgent(obj)


def cell(obj, value=None):
    """
    Returns a `cell`-agent of `obj`. This function can be used to manipulate
    the cell containers of `obj`.

    Args:
        obj (Reactive): A reactive object.

    Returns:
        An agent object offering the same attributes as `obj`, with the cell
        containers as attribute values.
    """
    if isinstance(obj, Pointer):
        holder, cell = _get_holder_cell(obj)
        if value is not None:
            if isinstance(value, Container):
                cell.set_container(holder, value)
            else:
                raise TypeError("must be of type container", value)

        return cell.get_container(holder)
    else:
        return CellAgent(obj)



import sys
from .pointer import SELF
# __pragma__ ('skip')
from .core import CellBase, Container, ResetContainer

__all__ = ("Cell", "MakeCell", "ResetCell", "TypeCell", "MakeTypeCell")
# __pragma__ ('noskip')
# __pragma__ ('ecom')
"""?
from .pcore import CellBase, Container, ResetContainer
?"""
# __pragma__ ('noecom')

_cell_counter = 0


class Cell(CellBase):
    """
    A descriptor to assign cell containers to objects.

    Args:
        default_val: The default value of the cell.
    """
    __container__ = Container  # factory
    __slots__ = ("order", "default")

    def __init__(self, default=None):
        super().__init__()
        global _cell_counter
        self.default = default
        self.order = _cell_counter
        _cell_counter += 1

    def create_container(self, holder):
        return self.__container__(self.default)

    def copy(self):
        r = self.__class__(self.default)
        r.order = self.order
        r.name = self.name
        r.index = self.index
        return r

    def __repr__(self):
        return "<{} {}({}/{})>".format(
            self.__class__.__name__, self.name, self.index,
            self.order)  # pragma: no cover


class MakeCell(Cell):
    """
    A cell that creates the default value with a factory during
    construction of the reactive object.

    Args:
        factory (callable): A factory that uses args and kwargs to
            create the default object.
        *args: argument list for factory
        **kwargs: keyword arguments for factory

    Examples:
        >>> @reactive
        class ListContainer(object):
            values = MakeCell(list, (1, 2, 3))
    """
    __slots__ = ("args", "kwargs")

    def __init__(self, factory, *args, **kwargs):
        super().__init__(factory)
        self.args = args
        self.kwargs = kwargs

    def copy(self):
        r = self.__class__(self.default, *self.args, **self.kwargs)
        r.order = self.order
        r.name = self.name
        r.index = self.index
        return r

    def create_container(self, holder):
        def replace_self(a):
            return holder if a is SELF else a

        args = [replace_self(a) for a in self.args]
        kwargs = dict((k, replace_self(v)) for k, v in self.kwargs.items())
        return Container(self.default(*args, **kwargs))


class ResetCell(Cell):
    """
    A Cell that silently resets it's value after all depending
    rules have been called.
    """
    __container__ = ResetContainer  # factory


class _TypeConverter(object):
    __slots__ = ("converter", )

    def __init__(self, converter):
        self.converter = converter

    def __call__(self, to_convert):
        converted = self.converter(to_convert)
        try:
            self.converter = converted.convert
        except AttributeError:
            pass

        return converted


class TypeCell(Cell):
    """
    A cell the coerces an input cell to the type of its default value.

    Args:
        default_val: The default value of the cell.
        converter (callable): If not None, the converter will be used to
            to coerce any value to the right type. Otherwise
            `type(default_val)` will be used as converter.
    """
    __slots__ = ("converter", )

    def __init__(self, default_val, converter=None):
        super().__init__(default_val)
        if converter is not None:
            self.converter = converter

        elif hasattr(getattr(default_val, "convert", None), '__call__'):
            self.converter = _TypeConverter(default_val.convert)

        else:
            self.converter = type(default_val)

    def __set__(self, holder, value):
        try:
            super().__set__(holder, self.converter(value))
        except BaseException:
            type_, exc, traceback = sys.exc_info()
            exc.args += (self,)
            if sys.version_info[0] >= 3:
                raise exc.with_traceback(traceback)
            else:  # pragma: no cover
                exec("raise exc, None, traceback")


class MakeTypeCell(MakeCell):
    def __set__(self, holder, value):
        super().__set__(holder, self.default(value))



from __future__ import print_function
import sys
import larch.reactive as ra

import time
import cProfile



class RTest(ra.Reactive):
    cell1 = ra.Cell(10)
    cell2 = ra.Cell(11)

    """
    @ra.rule
    def check_cell1(self):
        assert(self.cell1 < 1000)
    """

rinstances = [ RTest() for i in range(1000) ]

def rread():
    for i in range(100):
        for t in rinstances:
            a = t.cell1
            b = t.cell2


def rwrite():
    for i in range(100):
        for t in rinstances:
            t.cell1 = i
            t.cell2 = i



class Test(object):
    def __init__(self):
        self.cell1 = 10
        self.cell2 = 11


instances = [ Test() for i in range(1000) ]

def read():
    for i in range(100):
        for t in instances:
            a = t.cell1
            b = t.cell2


def write():
    for i in range(100):
        for t in instances:
            t.cell1 = i
            t.cell2 = i


if 1:
    print("Original Objects")
    print("================")
    start = time.time()
    read()
    delta = (time.time() - start) * 1000
    print("read test", delta)


    start = time.time()
    write()
    delta = (time.time() - start) * 1000
    print("write test", delta)




print("Reactive Objects")
print("================")
if 1:
    start = time.time()
    rread()
    delta = (time.time() - start) * 1000
    print("read test", delta)

    start = time.time()
    rwrite()
    delta = (time.time() - start) * 1000
    print("write test", delta)


cProfile.run('rread()')



import unittest
import gevent
import gevent.event as gev
import larch.reactive as ra
import logconfig


class ShutDownExecution(BaseException):
    def __init__(self, calling):
        self.calling = calling


class Test(ra.Reactive):
    tick = ra.Cell(0)
    no_touch = ra.Cell(0)  # not touched by any rule
    set_event = ra.Cell(None)
    wait_for_kill = ra.Cell(0)
    second_rule = ra.Cell(0)
    throw_greenlet = ra.Cell(None)

    def __init__(self, set_event=None):
        self.set_event = set_event
        self.calls = []

    @ra.rule
    def update_tick(self):
        if self.tick:
            self.calls.append(("before", gevent.getcurrent()))
            gevent.sleep(0)
            self.calls.append(("after", gevent.getcurrent()))

    @ra.rule
    def _rule_set_event(self):
        if self.set_event:
            self.set_event.set()
            gevent.sleep(0)

    @ra.rule
    def _rule_wait_for_kill(self):
        if self.wait_for_kill:
            self.second_rule = 1
            gevent.sleep(1000)

    @ra.rule
    def _rule_second(self):
        if self.second_rule:
            self.calls.append("called second rule")

            def callback():
                self.calls.append("called callback")

            ra.call_outside(callback)

    @ra.rule
    def _rule_thow_exception(self):
        g = self.throw_greenlet
        if g is not None:
            g.throw(ShutDownExecution(gevent.getcurrent()))


class TestLock(ra.Reactive):
    lock1 = ra.Cell(0)
    lock2 = ra.Cell(0)
    lock3 = ra.Cell(0)

    @ra.rule
    def _rule_lock_src(self):
        self.lock1
        yield
        gevent.sleep(0)  # change to greenlet 2
        self.lock2 = self.lock1  # new atomic action may not lock

    @ra.rule
    def _rule_lock_cause(self):
        self.lock3
        gevent.sleep(0)  # change back to greenlet1 atomic count=1

    @ra.rule
    def _rule_lock2(self):
        self.lock2  # touch


class GeventTest(unittest.TestCase):
    def test_event(self):
        event = gev.Event()
        started = [False]

        def run():
            event.wait()
            started[0] = True

        greenlet = gevent.spawn(run)
        gevent.sleep(0)
        t = Test()
        t.set_event = event

        greenlet.join()
        self.assertTrue(started[0])

    def test_kill_rule(self):
        exit_called = []
        t = Test()

        def activate_wait_rule():
            try:
                t.wait_for_kill = 1
            except gevent.GreenletExit:
                exit_called.append(1)

        logconfig.logstream.truncate(0)
        g = gevent.spawn(activate_wait_rule)
        gevent.sleep(0)
        g.kill()
        g.join()
        self.assertTrue(g.ready())
        self.assertEqual(t.calls, ['called second rule', 'called callback'])
        self.assertEqual(exit_called, [1])
        self.assertIn("GreenletExit during observer notification",
                      logconfig.logstream.getvalue())

    def test_kill_callback(self):
        calls = []

        def prolog():
            calls.append("prolog")

        def sleep():
            calls.append("before sleep")
            gevent.sleep(100)
            calls.append("after sleep")

        def epilog():
            calls.append("epilog")

        def activate_callbacks():
            with ra.atomic():
                ra.call_outside(prolog)
                ra.call_outside(sleep)
                ra.call_outside(epilog)

        logconfig.logstream.truncate(0)
        g = gevent.spawn(activate_callbacks)
        gevent.sleep(0)
        g.kill()
        g.join()
        self.assertTrue(g.ready())
        self.assertEqual(calls, ['prolog', 'before sleep', 'epilog'])
        self.assertIn("GreenletExit while executing callback",
                      logconfig.logstream.getvalue())

    def test_rule_throw_exception(self):
        t1 = Test()
        t2 = Test()

        def do_throw(g):
            t2.throw_greenlet = g

        def tick():
            gevent.spawn(do_throw, gevent.getcurrent())
            try:
                t1.tick = 1
                self.assertFalse(True)
            except ShutDownExecution:
                self.assertFalse(False)

        gevent.spawn(tick)

    def test_lock(self):
        # this constellation caused an eternal jump in the gevent loop

        t = TestLock()

        def call_lock1():
            t.lock1 = 1

        def call_lock2():
            t.lock3 = 2

        g1 = gevent.spawn(call_lock1)
        g2 = gevent.spawn(call_lock2)

        gevent.joinall([g1, g2])


if __name__ == "__main__":
    unittest.main()



# for ccore.pyx to ensure that decorated functions are pure python functions
def _make_handler(context, func):
    def handler(*args, **kwargs):
        return context.call(func, *args, **kwargs)

    return handler



import os
import platform
try:
    if platform.python_implementation() != "CPython":
        raise ImportError()

    if os.environ.get("LARCH_REACTIVE", "cython") != "cython":
        raise ImportError()

    from .ccore import (
        rcontext, CellBase, ReactiveState, Subject, Container, ResetContainer, Rule, IterRule,
        pointer_attrgetter, pointer_itemgetter, pointer_resolve, ReactiveWarning)
    core = "c"
except ImportError:
    from .pcore import (
        rcontext, CellBase, ReactiveState, Subject, Container, ResetContainer, Rule, IterRule,
        pointer_attrgetter, pointer_itemgetter, pointer_resolve, ReactiveWarning)
    core = "python"

__all__ = (
    "rcontext", "CellBase", "ReactiveState", "Subject", "Container", "ResetContainer", "Rule",
    "IterRule", "pointer_attrgetter", "pointer_itemgetter", "pointer_resolve", "ReactiveWarning",
    "core")



# !!! sdist does not add the needed c-headers
# don't blame me this is a bug of setuptools:
# https://github.com/pypa/setuptools/issues/1162
import os
from setuptools import setup, find_namespace_packages
from distutils.extension import Extension

debug = os.environ.get("LARCH_INSTALL_FOR_TEST")

try:
    from Cython.Distutils import build_ext

    try:
        # regenerate cython output file each time (see PY_VERSION_HEX)
        os.remove(os.path.join('larch', 'reactive', 'ccore.cpp'))
    except OSError:
        pass

    cdir = os.path.join('larch', 'reactive')
    path = os.path.join(cdir, 'ccore.pyx')
    ccore_extension = Extension('larch.reactive.ccore', [path])
    ccore_extension.language = "c++"
    ccore_extension.depends = [os.path.join(cdir, 'pyref.h'),
                               os.path.join(cdir, 'greenlet.h')]
    ccore_extension.cython_directives = {
        "boundscheck": False, "always_allow_keywords": False,
        "wraparound": False, "language_level": 3}

    if debug:
        ccore_extension.cython_directives.update({
            "linetrace": True, "profile": True})
        ccore_extension.define_macros.extend([
            ("CYTHON_TRACE_NOGIL", 1),
            ("CYTHON_TRACE", 1),
            ("DEBUG", None)])
        ccore_extension.extra_compile_args.extend(["-g"])

    # some setup tool bug changes ccore.pyx to ccore.c
    # update it again
    ccore_extension.sources = [path]
    ext_args = dict(
        cmdclass={'build_ext': build_ext}, ext_modules=[ccore_extension])

except ImportError:
    from distutils.command import build_ext
    path = os.path.join('larch', 'reactive', 'ccore.cpp')
    ccore_extension = Extension('larch.reactive.ccore', [path])
    ccore_extension.sources = [path]
    ext_args = dict(ext_modules=[ccore_extension])


setup(
    name='larch-reactive',
    version="4.0.12",
    packages=find_namespace_packages(where="./", include=["larch.reactive"]),
    zip_safe=False,
    extras_require={
        'documentation': ['sphinxcontrib-napoleon', 'sphinxcontrib-plantuml'],
        "build": ["cython"]
    },

    # metadata for upload to PyPI
    author='Michael Reithinger',
    description='reactive programming for python',
    license='BSD',
    keywords='library',
    url='https://github.com/kochelmonster/larch-reactive', 

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules"],

    **ext_args
)



"""
Tests threading and greenlet issues
"""
import threading
import larch.reactive as ra
import unittest


class Test(ra.Reactive):
    tick = ra.ResetCell(False)
    counter = 0

    @ra.rule
    def update_tick(self):
        if self.tick:
            self.counter += 1


class TestThread1(threading.Thread):
    def __init__(self):
        super(TestThread1, self).__init__()

    def run(self):
        self.test_obj = Test()

        for i in range(10):
            self.test_obj.tick = True

        self.rounds = ra.rcontext.rounds


class TestThread2(threading.Thread):
    def __init__(self):
        super(TestThread2, self).__init__()

    def run(self):
        self.test_obj = Test()

        for i in range(20):
            self.test_obj.tick = True

        self.rounds = ra.rcontext.rounds


class ThreadTest(unittest.TestCase):
    def test_threading(self):
        rounds = ra.rcontext.rounds
        t1 = TestThread1()
        t2 = TestThread2()

        t1.start()
        t2.start()
        t1.join()
        t2.join()

        self.assertNotEqual(t1.rounds, t2.rounds)
        self.assertEqual(t1.test_obj.counter, 10)
        self.assertEqual(t2.test_obj.counter, 20)
        self.assertEqual(ra.rcontext.rounds, rounds)


if __name__ == "__main__":
    unittest.main()
